import os
import re
import subprocess

import warnings


# %%
def get_core_count():
    """Get the number of available virtual or physical CPUs on this system"""

    # cpuset
    # cpuset may restrict the number of *available* processors
    try:
        m = re.search(r"(?m)^Cpus_allowed:\s*(.*)$", open("/proc/self/status").read())
        if m:
            res = bin(int(m.group(1).replace(",", ""), 16)).count("1")
            if res > 0:
                return res
    except IOError:
        pass

    # Python 2.6+
    try:
        import multiprocessing

        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    # https://github.com/giampaolo/psutil
    try:
        import psutil

        return psutil.cpu_count()  # psutil.NUM_CPUS on old versions
    except (ImportError, AttributeError):
        pass

    # POSIX
    try:
        res = int(os.sysconf("SC_NPROCESSORS_ONLN"))

        if res > 0:
            return res
    except (AttributeError, ValueError):
        pass

    # Windows
    try:
        res = int(os.environ["NUMBER_OF_PROCESSORS"])

        if res > 0:
            return res
    except (KeyError, ValueError):
        pass

    # jython
    try:
        from java.lang import Runtime

        runtime = Runtime.getRuntime()
        res = runtime.availableProcessors()
        if res > 0:
            return res
    except ImportError:
        pass

    # BSD
    try:
        sysctl = subprocess.Popen(["sysctl", "-n", "hw.ncpu"], stdout=subprocess.PIPE)
        scStdout = sysctl.communicate()[0]
        res = int(scStdout)

        if res > 0:
            return res
    except (OSError, ValueError):
        pass

    # Linux
    try:
        res = open("/proc/cpuinfo").read().count("processor\t:")

        if res > 0:
            return res
    except IOError:
        pass

    # Solaris
    try:
        pseudoDevices = os.listdir("/devices/pseudo/")
        res = 0
        for pd in pseudoDevices:
            if re.match(r"^cpuid@[0-9]+$", pd):
                res += 1

        if res > 0:
            return res
    except OSError:
        pass

    # Other UNIXes (heuristic)
    try:
        try:
            dmesg = open("/var/run/dmesg.boot").read()
        except IOError:
            dmesgProcess = subprocess.Popen(["dmesg"], stdout=subprocess.PIPE)
            dmesg = dmesgProcess.communicate()[0]

        res = 0
        while "\ncpu" + str(res) + ":" in dmesg:
            res += 1

        if res > 0:
            return res
    except OSError:
        pass

    warnings.warn("Can not determine number of CPUs on this system", RuntimeWarning)



import pandas as pd
import os
from crabnet.utils.composition import _fractional_composition


def norm_form(formula):
    comp = _fractional_composition(formula)
    form = ""
    for key, value in comp.items():
        form += f"{key}{str(value)[0:9]}"
    return form


def count_elems(string):
    count = 0
    switch = 1
    for c in string:
        if c.isalpha():
            count += switch
            switch = 0
        if c.isnumeric():
            switch = 1
    return count


# %%
if __name__ == "__main__":
    print("processing all model predictions and calculating metrics")
    print("this will take a few minutes...")

    # %%
    results_path = "publication_predictions"
    benchmark_path = "data/benchmark_data"
    test_directories = os.listdir(results_path)
    benchmark_props = os.listdir(benchmark_path)

    benchmark_test_directories = [
        test for test in test_directories if "benchmark" in test
    ]

    dataset_results = {}
    dataset_preds = {}
    dataset_acts = {}
    test_maes = pd.DataFrame()
    df_stats = pd.DataFrame()
    for benchmark in benchmark_props:
        df_compositions = pd.DataFrame()
        df_preds = pd.DataFrame()
        df_acts = pd.DataFrame()
        models = []
        for directory in benchmark_test_directories:
            df_train_orig = pd.read_csv(
                f"{benchmark_path}/{benchmark}/train.csv",
                keep_default_na=False,
                na_values=[""],
            )
            df_val = pd.read_csv(
                f"{benchmark_path}/{benchmark}/val.csv",
                keep_default_na=False,
                na_values=[""],
            )
            df_train = pd.concat([df_train_orig, df_val], ignore_index=True)
            df_train["formula"] = [
                norm_form(formula) for formula in df_train["formula"]
            ]
            df_train.index = df_train["formula"]
            files = os.listdir(f"{results_path}\{directory}")
            file = [file for file in files if benchmark in file and "test" in file]
            if len(file) > 0:
                models.append(directory.split("_")[0])
                file = file[0]
                df = pd.read_csv(
                    f"{results_path}\{directory}\{file}",
                    keep_default_na=False,
                    na_values=[""],
                )
                composition = df["formula"]
                pred = df["predicted"]
                act = df["actual"]
                print(f"processing {benchmark} {models[-1]}")
                df_compositions = pd.concat([df_compositions, composition], axis=1)
                df_preds = pd.concat([df_preds, pred], axis=1)
                df_acts = pd.concat([df_acts, act], axis=1)

                n_total = act.count() + df_val.shape[0] + df_train_orig.shape[0]
                df_stats.at[benchmark, "mean_test"] = act.mean()
                df_stats.at[benchmark, "std_test"] = act.std()
                df_stats.at[benchmark, "n_test"] = act.count()

                df_stats.at[benchmark, "mean_train"] = df_train["target"].mean()
                df_stats.at[benchmark, "std_train"] = df_train["target"].std()
                df_stats.at[benchmark, "n_train"] = df_train_orig.shape[0]

                df_stats.at[benchmark, "n_val"] = df_val.shape[0]
                df_stats.at[benchmark, "n_total"] = n_total

                df_stats.at[benchmark, "prop_train"] = df_train_orig.shape[0] / n_total
                df_stats.at[benchmark, "prop_val"] = df_val.shape[0] / n_total
                df_stats.at[benchmark, "prop_test"] = act.count() / n_total

        df_compositions.columns = models
        df_preds.columns = models
        df_acts.columns = models
        df_diff = df_preds - df_acts
        df_mae = df_diff.abs().mean()
        test_maes[benchmark] = df_mae

        dataset_results[benchmark] = df_compositions
        dataset_preds[benchmark] = df_preds
        dataset_acts[benchmark] = df_acts

    maes = test_maes.T
    model_names = ["roost", "mat2vec", "onehot", "elemnet", "rf"]
    out_1 = maes[model_names]
    out = pd.concat([out_1, df_stats], axis=1)
    df_benchmark = out.copy()

    # %%
    results_path = "publication_predictions"
    matbench_path = "data/matbench_cv"
    test_directories = os.listdir(results_path)
    matbench_props = os.listdir(matbench_path)

    matbench_test_directories = [
        test for test in test_directories if "matbench" in test
    ]

    dataset_results = {}
    dataset_preds = {}
    dataset_acts = {}
    test_maes = pd.DataFrame()
    df_stats = pd.DataFrame()
    for matbench in matbench_props:
        df_compositions = pd.DataFrame()
        df_preds = pd.DataFrame()
        df_acts = pd.DataFrame()
        models = []
        for directory in matbench_test_directories:
            train_files = os.listdir(f"{matbench_path}/{matbench}")
            train_files = [file for file in train_files if "train" in file]
            test_files = os.listdir(f"{results_path}/{directory}")
            test_files = [
                file for file in test_files if matbench in file and "test" in file
            ]
            for i, (train_file, test_file) in enumerate(zip(train_files, test_files)):
                df_train_orig = pd.read_csv(
                    f"{matbench_path}/{matbench}/{train_file}",
                    keep_default_na=False,
                    na_values=[""],
                )
                df_val = pd.read_csv(
                    f'{matbench_path}/{matbench}/{train_file.replace("train", "val")}',
                    keep_default_na=False,
                    na_values=[""],
                )
                df_train = pd.concat([df_train_orig, df_val], ignore_index=True)
                df_train["formula"] = [
                    norm_form(formula) for formula in df_train["formula"]
                ]
                df_train.index = df_train["formula"]
                if len(file) > 0:
                    models.append(directory.split("_")[0] + f"_{i}")
                    file = file[0]
                    df = pd.read_csv(
                        f"{results_path}\{directory}\{test_file}",
                        keep_default_na=False,
                        na_values=[""],
                    )
                    df.index = df["formula"].values
                    composition = df["formula"]
                    pred = df["predicted"]
                    act = df["actual"]
                    print(f"processing {matbench} {models[-1]}")
                    df_compositions = pd.concat([df_compositions, composition], axis=1)
                    df_preds = pd.concat([df_preds, pred], axis=1)
                    df_acts = pd.concat([df_acts, act], axis=1)

                    n_total = act.count() + df_val.shape[0] + df_train_orig.shape[0]
                    df_stats.at[matbench, "mean_test"] = act.mean()
                    df_stats.at[matbench, "std_test"] = act.std()
                    df_stats.at[matbench, "n_test"] = act.count()

                    df_stats.at[matbench, "mean_train"] = df_train["target"].mean()
                    df_stats.at[matbench, "std_train"] = df_train["target"].std()
                    df_stats.at[matbench, "n_train"] = df_train_orig.shape[0]

                    df_stats.at[matbench, "n_val"] = df_val.shape[0]
                    df_stats.at[matbench, "n_total"] = n_total

                    df_stats.at[matbench, "prop_train"] = (
                        df_train_orig.shape[0] / n_total
                    )
                    df_stats.at[matbench, "prop_val"] = df_val.shape[0] / n_total
                    df_stats.at[matbench, "prop_test"] = act.count() / n_total

        df_compositions.columns = models
        df_preds.columns = models
        df_acts.columns = models
        df_diff = df_preds - df_acts
        df_mae_cv = df_diff.abs()
        df_mae = pd.DataFrame()
        model_names = []
        _ = [model_names.append(x[:-2]) for x in models if x[:-2] not in model_names]
        for j, model in enumerate(model_names):
            df_mae.loc[model, 0] = (
                df_mae_cv.iloc[:, (j) * 5 : (j + 1) * 5].max(axis=1).mean()
            )

        test_maes[matbench] = df_mae[0]
        dataset_results[matbench] = df_compositions
        dataset_preds[matbench] = df_preds
        dataset_acts[matbench] = df_acts

    maes = test_maes.T
    model_names = ["roost", "mat2vec", "onehot", "elemnet", "rf"]
    out_1 = maes[model_names]
    out = pd.concat([out_1, df_stats], axis=1)
    df_matbench = out.copy()

    # %%
    df_all = pd.concat([df_benchmark, df_matbench], axis=0, ignore_index=False)

    rename_dict = {"mat2vec": "crabnet", "onehot": "hotcrab"}
    df_all = df_all.rename(columns=rename_dict)
    df_all.to_csv("metrics/all_metrics.csv", index_label="property")



import os
import numpy as np
import pandas as pd

from tqdm import tqdm
from crabnet.utils.composition import _element_composition


# %%
benchmark_data = os.listdir("data/benchmark_data")
matbench_data = os.listdir("data/matbench_cv")


# %%
cols = [
    "mat_prop",
    "samples_total",
    "samples_train",
    "samples_val",
    "samples_test",
    "prop_train",
    "prop_val",
    "prop_test",
]
df_benchmark = pd.DataFrame(columns=cols)

for prop in tqdm(benchmark_data, desc="Processing benchmark_data"):
    df_train = pd.read_csv(
        f"data/benchmark_data/{prop}/train.csv", keep_default_na=False, na_values=[""]
    )
    df_val = pd.read_csv(
        f"data/benchmark_data/{prop}/val.csv", keep_default_na=False, na_values=[""]
    )
    df_test = pd.read_csv(
        f"data/benchmark_data/{prop}/test.csv", keep_default_na=False, na_values=[""]
    )
    n_train = df_train["formula"].apply(_element_composition).apply(len).max()
    n_val = df_val["formula"].apply(_element_composition).apply(len).max()
    n_test = df_test["formula"].apply(_element_composition).apply(len).max()
    n_elements = max([n_train, n_val, n_test])
    total_samples = df_train.shape[0] + df_val.shape[0] + df_test.shape[0]
    df_row = {
        "mat_prop": prop,
        "n_elements": n_elements,
        "samples_total": total_samples,
        "samples_train": df_train.shape[0],
        "samples_val": df_val.shape[0],
        "samples_test": df_test.shape[0],
        "prop_train": df_train.shape[0] / total_samples,
        "prop_val": df_val.shape[0] / total_samples,
        "prop_test": df_test.shape[0] / total_samples,
    }
    df_benchmark = df_benchmark.append(df_row, ignore_index=True)


# %%
cols = [
    "mat_prop",
    "samples_total",
    "samples_train",
    "samples_val",
    "samples_test",
    "prop_train",
    "prop_val",
    "prop_test",
]
df_matbench_cv = pd.DataFrame(columns=cols)

for prop in tqdm(matbench_data, desc="Processing matbench_data"):
    df_train = pd.read_csv(
        f"data/matbench_cv/{prop}/train0.csv", keep_default_na=False, na_values=[""]
    )
    df_val = pd.read_csv(
        f"data/matbench_cv/{prop}/val0.csv", keep_default_na=False, na_values=[""]
    )
    df_test = pd.read_csv(
        f"data/matbench_cv/{prop}/test0.csv", keep_default_na=False, na_values=[""]
    )
    n_train = df_train["formula"].apply(_element_composition).apply(len).max()
    n_val = df_val["formula"].apply(_element_composition).apply(len).max()
    n_test = df_test["formula"].apply(_element_composition).apply(len).max()
    n_elements = max([n_train, n_val, n_test])
    total_samples = df_train.shape[0] + df_val.shape[0] + df_test.shape[0]
    df_row = {
        "mat_prop": prop,
        "n_elements": n_elements,
        "samples_total": total_samples,
        "samples_train": df_train.shape[0],
        "samples_val": df_val.shape[0],
        "samples_test": df_test.shape[0],
        "prop_train": df_train.shape[0] / total_samples,
        "prop_val": df_val.shape[0] / total_samples,
        "prop_test": df_test.shape[0] / total_samples,
    }
    df_matbench_cv = df_matbench_cv.append(df_row, ignore_index=True)


# %%
# cols = ['mat_prop', 'samples_total']
# df_matbench_all = pd.DataFrame(columns=cols)

# for prop in tqdm(matbench_data, desc="Processing matbench_data"):
#     df_data = pd.read_csv(f'data/matbench/{prop}.csv',
#                           keep_default_na=False, na_values=[''])
#     n_elements = df_data['formula'].apply(_element_composition).apply(len).max()
#     df_row = {
#         'mat_prop': prop,
#         'n_elements': n_elements,
#         'samples_total': df_data.shape[0],
#         }
#     df_matbench_all = df_matbench_all.append(df_row, ignore_index=True)


# %%
df_benchmark["log2_samples_train"] = (
    df_benchmark["samples_train"].astype(float).apply(np.log2)
)
df_matbench_cv["log2_samples_train"] = (
    df_matbench_cv["samples_train"].astype(float).apply(np.log2)
)
# df_matbench_all['log2_samples_total'] = df_matbench_all['samples_total'].astype(float).apply(np.log2)


# %%
print(df_benchmark.to_latex(index=False, escape=True, float_format="{:0.2f}".format))

# %%
print(df_matbench_cv.to_latex(index=False, escape=True, float_format="{:0.2f}".format))

# %%
# print(df_matbench_all.to_latex(index=False, escape=True, float_format="{:0.2f}".format))


# %%
df_benchmark
df_matbench_cv
# df_matbench_all


# %%
df_benchmark.to_csv("df_benchmark.csv")
df_matbench_cv.to_csv("df_matbench_cv.csv")



"""Test CrabNet's fit and predict via `get_model()` and `predict()`."""
from crabnet.model import data
from crabnet.data.materials_data import elasticity
from crabnet.train_crabnet import get_model


def test_crabnet():
    train_df, val_df = data(elasticity)
    mdl = get_model(train_df=train_df, val_df=val_df)

    train_true, train_pred, formulas, train_sigma = mdl.predict(val_df)
    return train_true, train_pred, formulas, train_sigma


if __name__ == "__main__":
    train_true, train_pred, formulas, train_sigma = test_crabnet()



import os
import numpy as np
import pandas as pd
import seaborn as sns
from tqdm import tqdm

import torch

from sklearn.metrics import roc_auc_score

from crabnet.kingcrab import CrabNet
from crabnet.model import Model
from crabnet.utils.get_compute_device import get_compute_device

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.cm as cm
from matplotlib.colors import Normalize

compute_device = get_compute_device()
data_type = torch.float


# %%
# fmt: off
all_symbols = ['None', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na',
               'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc',
               'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
               'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb',
               'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb',
               'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
               'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
               'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl',
               'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
               'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
               'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg',
               'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
# fmt: on


def model(mat_prop, classification_list, simple=False):
    # Get the TorchedCrabNet architecture loaded
    model = Model(
        CrabNet(compute_device=compute_device).to(compute_device),
        model_name=f"{mat_prop}",
    )
    if True:
        model.load_network(f"{mat_prop}.pth")
        model.model_name = f"{mat_prop}"

    if mat_prop in classification_list:
        model.classification = True

    dataset = rf"{data_dir}\{mat_prop}\train.csv"
    model.load_data(dataset, batch_size=2 ** 7)  # data is reloaded to model.data_loader

    model.model.eval()
    model.model.avg = False

    simple_tracker = {i: [] for i in range(119)}
    element_tracker = {i: [] for i in range(119)}
    composition_tracker = {}

    with torch.no_grad():
        for i, data in enumerate(tqdm(model.data_loader)):
            X, y, formula = data
            src, frac = X.squeeze(-1).chunk(2, dim=1)
            src = src.to(compute_device, dtype=torch.long, non_blocking=True)
            frac = frac.to(compute_device, dtype=data_type, non_blocking=True)
            y = y.to(compute_device, dtype=data_type, non_blocking=True)
            output = model.model.forward(src, frac)
            mask = (src == 0).unsqueeze(-1).repeat(1, 1, 1)
            prediction, uncertainty, prob = output.chunk(3, dim=-1)
            prediction = prediction * torch.sigmoid(prob)
            uncertainty = torch.exp(uncertainty) * model.scaler.std
            prediction = model.scaler.unscale(prediction)
            prediction = prediction * ~mask
            uncertainty = uncertainty * ~mask
            if model.classification:
                prediction = torch.sigmoid(prediction)
            for i in range(src.shape[0]):
                if any(prediction[i].cpu().numpy().ravel() < 0):
                    composition_tracker[formula[i]] = [
                        src[i].cpu().numpy(),
                        frac[i].cpu().numpy(),
                        y[i].cpu().numpy(),
                        prediction[i].cpu().numpy(),
                        uncertainty[i].cpu().numpy(),
                    ]
                for j in range(src.shape[1]):
                    element_tracker[int(src[i][j])].append(float(prediction[i][j]))
                    simple_tracker[int(src[i][j])].append(float(y[i]))

    def elem_view(element_tracker, plot=True):
        property_tracker = {}
        x_max = max([y[1] for y in model.data_loader.dataset])
        x_min = min([y[1] for y in model.data_loader.dataset])
        x_range = x_max - x_min
        x_min_buffer = 0.1 * x_range
        x_max_buffer = 0.1 * x_range
        for key in element_tracker.keys():
            data = element_tracker[key]
            if len(data) > 10:
                sum_prop = sum(data)
                mean_prop = sum_prop / len(data)
                prop = mean_prop
                property_tracker[all_symbols[key]] = prop
                if plot:
                    plt.figure(figsize=(4, 4))
                    hist_kws = {
                        "edgecolor": "k",
                        "linewidth": 2,
                        "alpha": 1,
                        "facecolor": "#A1D884",
                    }
                    ax = sns.distplot(
                        data,
                        label=f"{all_symbols[key]}, n={len(data)}",
                        kde=False,
                        bins=np.arange(0, 500, 25),
                        hist_kws=hist_kws,
                        kde_kws={"color": "k", "linewidth": 2},
                    )

                    ax.axes.yaxis.set_visible(False)
                    plt.legend()
                    plt.xlim(x_min - x_min_buffer, x_max + x_max_buffer)
                    plt.xlabel("Bulk Modulus Contribution (GPa)")
                    plt.tick_params(axis="both", which="both", direction="in")

                    save_dir = f"figures/contributions/{mat_prop}/"
                    os.makedirs(save_dir, exist_ok=True)
                    plt.savefig(
                        f"{save_dir}{all_symbols[key]}.png",
                        dpi=300,
                        bbox_inches="tight",
                    )
                    plt.show()
        return property_tracker

    if simple:
        property_tracker = elem_view(simple_tracker, plot=True)
    else:
        property_tracker = elem_view(element_tracker, plot=True)

    return property_tracker


def save_results(output, save_name):
    act, pred, formulae, uncertainty = output
    df = pd.DataFrame([formulae, act, pred, uncertainty]).T
    df.columns = ["composition", "target", "pred-0", "uncertainty"]
    save_path = "new_results"


def save_test_results(mat_prop, classification_list):
    # Load up a saved network.
    model = Model(CrabNet(compute_device=compute_device).to(compute_device))
    model.load_network(f"{mat_prop}.pth")
    if mat_prop in classification_list:
        model.classification = True
    # Load the data you want to predict with
    test_data = rf"data\benchmark_data\{mat_prop}\test.csv"
    model.load_data(test_data)  # data is reloaded to model.data_loader
    output = model.predict(model.data_loader)  # predict the data saved here
    if model.classification:
        auc = roc_auc_score(output[0], output[1])
        print(f"\n{mat_prop} ROC AUC: {auc:0.3f}")
    else:
        print(f"\n{mat_prop} mae: {abs(output[0] - output[1]).mean():0.3f}")
    # save your predictions to a csv
    save_results(output, f"{mat_prop}_output.csv")


# %%
def plot(mat_prop, property_tracker):
    ptable = pd.read_csv("data/element_properties/ptable.csv")
    ptable.index = ptable["symbol"].values
    elem_tracker = ptable["count"]
    n_row = ptable["row"].max()
    n_column = ptable["column"].max()

    elem_tracker = elem_tracker + pd.Series(property_tracker).drop("None", axis=0)

    # log_scale = True
    log_scale = False

    fig, ax = plt.subplots(figsize=(n_column, n_row))
    rows = ptable["row"]
    columns = ptable["column"]
    symbols = ptable["symbol"]
    rw = 0.9  # rectangle width (rw)
    rh = rw  # rectangle height (rh)
    for row, column, symbol in zip(rows, columns, symbols):
        row = ptable["row"].max() - row
        cmap = cm.YlGn
        count_min = 0
        count_max = elem_tracker.max() + 50
        norm = Normalize(vmin=count_min, vmax=count_max)
        count = elem_tracker[symbol]
        if log_scale:
            norm = Normalize(vmin=np.log(1), vmax=np.log(count_max))
            if count != 0:
                count = np.log(count)
        color = cmap(norm(count))
        if np.isnan(count):
            color = "silver"
        if row < 3:
            row += 0.5
        # element box
        rect = patches.Rectangle(
            (column, row),
            rw,
            rh,
            linewidth=1.5,
            edgecolor="gray",
            facecolor=color,
            alpha=1,
        )
        # plot element text
        plt.text(
            column + rw / 2,
            row + rw / 2,
            symbol,
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=22,
            fontweight="semibold",
            color="k",
        )

        ax.add_patch(rect)

    granularity = 20
    for i in range(granularity):
        value = (1 - i / (granularity - 1)) * count_min + (
            i / (granularity - 1)
        ) * count_max
        if log_scale:
            if value != 0:
                value = np.log(value)
        color = cmap(norm(value))
        length = 9
        x_offset = 3.5
        y_offset = 7.8
        x_loc = i / (granularity) * length + x_offset
        width = length / granularity
        height = 0.35
        rect = patches.Rectangle(
            (x_loc, y_offset),
            width,
            height,
            linewidth=1.5,
            edgecolor="gray",
            facecolor=color,
            alpha=1,
        )

        if i in [0, 4, 9, 14, 19]:
            text = f"{value:0.0f}"
            if log_scale:
                text = f"{np.exp(value):0.1e}".replace("+", "")
            plt.text(
                x_loc + width / 2,
                y_offset - 0.4,
                text,
                horizontalalignment="center",
                verticalalignment="center",
                fontweight="semibold",
                fontsize=20,
                color="k",
            )

        ax.add_patch(rect)

    name = "Bulk Modulus"
    legend_title = f"Average {name} Contribution"
    plt.text(
        x_offset + length / 2,
        y_offset + 0.7,
        f"log({legend_title})" if log_scale else legend_title,
        horizontalalignment="center",
        verticalalignment="center",
        fontweight="semibold",
        fontsize=20,
        color="k",
    )

    ax.set_ylim(-0.15, n_row + 0.1)
    ax.set_xlim(0.85, n_column + 1.1)

    # fig.patch.set_visible(False)
    ax.axis("off")

    plt.draw()
    save_dir = "figures/contributions"
    # save_dir = None
    if save_dir is not None:
        name = mat_prop
        fig_name = f"{save_dir}/{mat_prop}/ptable.png"
        os.makedirs(save_dir, exist_ok=True)
        plt.savefig(fig_name, bbox_inches="tight", dpi=300)

    plt.pause(0.001)
    plt.close()


# %%
data_dir = "data/benchmark_data"
matbench_props = os.listdir(data_dir)[0:1]

classification_list = []

mat_prop = "aflow__ael_bulk_modulus_vrh"
property_tracker = model(mat_prop, classification_list)
plot(mat_prop, property_tracker)



import os
import numpy as np
import pandas as pd
import torch

from sklearn.metrics import roc_auc_score

from crabnet.kingcrab import CrabNet
from crabnet.model import Model
from crabnet.utils.get_compute_device import get_compute_device

compute_device = get_compute_device(prefer_last=False)
RNG_SEED = 42
torch.manual_seed(RNG_SEED)
np.random.seed(RNG_SEED)


# %%
def get_model(
    mat_prop, i, classification=False, batch_size=None, transfer=None, verbose=True
):
    # Get the TorchedCrabNet architecture loaded
    model = Model(
        CrabNet(compute_device=compute_device).to(compute_device),
        model_name=f"{mat_prop}{i}",
        verbose=verbose,
    )

    # Train network starting at pretrained weights
    if transfer is not None:
        model.load_network(f"{transfer}.pth")
        model.model_name = f"{mat_prop}"

    # Apply BCEWithLogitsLoss to model output if binary classification is True
    if classification:
        model.classification = True

    # Get the datafiles you will learn from
    train_data = f"data/matbench_cv/{mat_prop}/train{i}.csv"
    val_data = f"data/matbench_cv/{mat_prop}/val{i}.csv"

    # Load the train and validation data before fitting the network
    data_size = pd.read_csv(train_data).shape[0]
    batch_size = 2 ** round(np.log2(data_size) - 4)
    if batch_size < 2 ** 7:
        batch_size = 2 ** 7
    if batch_size > 2 ** 12:
        batch_size = 2 ** 12
    # batch_size = 2**7
    model.load_data(train_data, batch_size=batch_size, train=True)
    print(
        f"training with batchsize {model.batch_size} "
        f"(2**{np.log2(model.batch_size):0.3f})"
    )
    model.load_data(val_data, batch_size=batch_size)

    # Set the number of epochs, decide if you want a loss curve to be plotted
    model.fit(epochs=300, losscurve=False)

    # Save the network (saved as f"{model_name}.pth")
    model.save_network()
    return model


def to_csv(output, save_name):
    # parse output and save to csv
    act, pred, formulae, uncertainty = output
    df = pd.DataFrame([formulae, act, pred, uncertainty]).T
    df.columns = ["formula", "actual", "predicted", "uncertainty"]
    save_path = "publication_predictions/mat2vec_matbench__predictions"
    # save_path = 'publication_predictions/onehot_matbench__predictions'
    # save_path = 'publication_predictions/random_200_matbench__predictions'
    os.makedirs(save_path, exist_ok=True)
    df.to_csv(f"{save_path}/{save_name}", index_label="Index")


def load_model(mat_prop, i, classification, file_name, verbose=True):
    # Load up a saved network.
    model = Model(
        CrabNet(compute_device=compute_device).to(compute_device),
        model_name=f"{mat_prop}{i}",
        verbose=verbose,
    )
    model.load_network(f"{mat_prop}{i}.pth")

    # Check if classifcation task
    if classification:
        model.classification = True

    # Load the data you want to predict with
    data = f"data/matbench_cv/{mat_prop}/{file_name}"
    # data is reloaded to model.data_loader
    model.load_data(data, batch_size=2 ** 9)
    return model


def get_results(model):
    output = model.predict(model.data_loader)  # predict the data saved here
    return model, output


def save_results(mat_prop, i, classification, file_name, verbose=True):
    model = load_model(mat_prop, i, classification, file_name, verbose=verbose)
    model, output = get_results(model)

    # Get appropriate metrics for saving to csv
    if model.classification:
        auc = roc_auc_score(output[0], output[1])
        print(f"\n{mat_prop}{i} ROC AUC: {auc:0.3f}")
    else:
        mae = np.abs(output[0] - output[1]).mean()
        print(f"\n{mat_prop}{i} mae: {mae:0.3g}")

    # save predictions to a csv
    fname = f'{mat_prop}_{file_name.replace(f"{i}.csv", "")}_output_cv{i}.csv'
    to_csv(output, fname)
    return model, mae


# %%
if __name__ == "__main__":
    # Get data to benchmark on
    data_dir = "data/matbench_cv"
    mat_props = os.listdir(data_dir)
    classification_list = []
    print(f"training: {mat_props}")
    for mat_prop in mat_props:
        classification = False
        if mat_prop in classification_list:
            classification = True
        # matbench provides 5 dataset train/val splits
        n_splits = 5
        maes = []
        for i in range(n_splits):
            print(f"property: {mat_prop}, cv {i}")
            model = get_model(mat_prop, i, classification, verbose=True)
            print("=====================================================")
            print("calculating test mae")
            model_test, t_mae = save_results(
                mat_prop, i, classification, f"test{i}.csv", verbose=False
            )
            print("calculating val mae")
            model_val, v_mae = save_results(
                mat_prop, i, classification, f"val{i}.csv", verbose=False
            )
            maes.append(t_mae)
        print(f"Average test mae: {np.mean(maes)}")
        print("=====================================================")



import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np

from crabnet.kingcrab import CrabNet
from crabnet.model import Model
from crabnet.utils.get_compute_device import get_compute_device
from benchmark_crabnet import get_results

import torch

from crabnet.utils.utils import CONSTANTS

compute_device = get_compute_device()


# %%
mat_prop = "mp_bulk_modulus"
crabnet_params = {"d_model": 512, "N": 3, "heads": 4}

model = Model(
    CrabNet(**crabnet_params, compute_device=compute_device).to(compute_device)
)
model.load_network(f"{mat_prop}.pth")

# Load the data you want to predict with
test_data = rf"data\benchmark_data\{mat_prop}\val.csv"
model.load_data(test_data)  # data is reloaded to model.data_loader
output = model.predict(model.data_loader)  # predict the data saved here


# %%
class SaveOutput:
    def __init__(self):
        self.outputs = []

    def __call__(self, module, module_in, module_out):
        self.outputs.append(module_out)

    def clear(self):
        self.outputs = []


save_output = SaveOutput()
hook_handles = []

for layer in model.model.modules():
    if isinstance(layer, torch.nn.modules.activation.MultiheadAttention):
        # print('isinstance')
        handle = layer.register_forward_hook(save_output)
        hook_handles.append(handle)

len(save_output.outputs)
model, _ = get_results(model)
len(save_output.outputs)


# %%
mod_out = save_output.outputs

len(mod_out)
len(mod_out[0])

n_mats = len(mod_out)  # number of output matrices from hook
bsz = model.data_loader.batch_size  # batch size from data loader
B = len(model.data_loader)  # total number of batches from data loader
H = model.model.heads  # number of heads
N = model.model.N  # number of layers
n_data = len(model.data_loader.dataset)
n_elements = model.n_elements

assert n_mats == N * B, "something is wrong with the matrices"

attn_data = torch.zeros(size=(n_data, N, H, n_elements, n_elements))
for layer in range(N):
    sliceN = [mod_out[i][1].unsqueeze(1) for i in range(layer, n_mats, N)]
    sliceN = torch.cat(sliceN, dim=0)
    attn_data[:, layer : layer + 1, :, :, :] = sliceN

attn_data = attn_data.detach().cpu().numpy()
data_loader = model.data_loader
train_loader = model.train_loader


# %%
def get_datum(data_loader, idx=0):
    datum = data_loader.dataset[idx]
    return datum


def get_x(data_loader, idx=0):
    x = get_datum(data_loader, idx=idx)[0]
    return x


def get_atomic_numbers(data_loader, idx=0):
    nums = get_x(data_loader, idx=idx).chunk(2)[0].detach().cpu().numpy()
    nums = nums.astype(int)
    return nums


def get_atomic_fracs(data_loader, idx=0):
    nums = get_x(data_loader, idx=idx).chunk(2)[1].detach().cpu().numpy()
    return nums


def get_target(data_loader, idx=0):
    target = get_datum(data_loader, idx=idx)[1].detach().cpu().numpy()
    return target


def get_form(data_loader, idx=0):
    form = get_datum(data_loader, idx=idx)[2]
    return form


# %%
def get_attention(attn_mat, idx=0, layer=0, head=0):
    """
    Get one slice of the attention map.

    Parameters
    ----------
    attn_mat : Tensor
        attn_mat is numpy array in the shape of [S, N, H, d, d], where
        S is the total number of data samples,
        N is the layer number in the attention mechanism,
        H is the head number in the attention mechanism, and
        d is the attention dimension in each head.
    idx : int, optional
        Index of the input material. The default is 0.
    layer : int, optional
        Layer number in the attention mechanism. The default is 0.
    head : int, optional
        Head number in the attention mechanism. The default is 0.

    Returns
    -------
    attn : Tensor

    """
    attn_mat = attn_mat
    assert len(attn_mat.shape) == 5, "input attn_map is of the wrong shape"
    attn = attn_mat[idx, layer, head, :, :]
    return attn


def plot_attention(
    map_data,
    cbar_ax=None,
    xlabel=None,
    ylabel=None,
    xticklabels=None,
    yticklabels=None,
    mask=True,
    ax=None,
):
    "plots ONE attention map slice = map_data on given axis"
    if mask is not None:
        map_data = map_data * mask
    if xticklabels is None:
        xticklabels = list(range(map_data.shape[0]))
    if yticklabels is None:
        yticklabels = list(range(map_data.shape[0]))
    xticklabels = [f"{i:0.3g}" for i in xticklabels]

    import matplotlib as mpl

    cmap1 = mpl.colors.ListedColormap(["lightgray"])

    ax = sns.heatmap(
        map_data,
        cmap=cmap1,
        mask=map_data != 0,
        linewidths=1,
        linecolor="w",
        cbar_ax=cbar_ax,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
        annot=True,
        fmt=".0f",
        annot_kws={"size": 15},
        vmin=0,
        vmax=0,
        ax=ax,
    )

    ax = sns.heatmap(
        map_data,
        cmap="rocket_r",
        mask=map_data == 0,
        linewidths=1,
        linecolor="gray",
        cbar_ax=cbar_ax,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
        annot=True,
        fmt=".2f",
        annot_kws={"size": 15},
        vmin=0,
        vmax=1,
        ax=ax,
    )

    for tick in ax.yaxis.get_major_ticks():
        tick.label1.set_verticalalignment("center")

    ax2 = ax.twiny()
    ax2.set_xlim(ax.get_xlim())
    ax2.set_xticks(ax.get_xticks())
    ax2.set_xticklabels(yticklabels)
    ax2.spines["left"].set_visible(False)
    ax2.spines["bottom"].set_visible(False)
    ax2.spines["right"].set_visible(False)
    ax2.spines["top"].set_visible(False)

    ax.xaxis.set_tick_params(
        bottom=False, top=False, right=False, left=False, labelbottom=True
    )
    ax2.xaxis.set_tick_params(bottom=False, top=False, right=False, left=False)
    ax.yaxis.set_tick_params(
        bottom=False, top=False, right=False, left=False, labelleft=True
    )
    ax2.yaxis.set_tick_params(bottom=False, top=False, right=False, left=False)

    ax.set_xticks([0.5, 1.5, 2.5, 3.5])
    ax.set_xticklabels(xticklabels)
    plt.setp(ax.get_xticklabels(), visible=True)

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)


def plot_all_heads(data_loader, attn_mat, idx=0, layer=0, mask=True):
    (S, N, H, d, d) = attn_mat.shape
    if H % 6 == 0:
        ncols = 6
    if H % 5 == 0:
        ncols = 5
    elif H % 4 == 0:
        if H == 4:
            ncols = 2
        else:
            ncols = 4
    elif H % 3 == 0:
        ncols = 3
    elif H % 2 == 0:
        ncols = 2
    else:
        ncols = 1
    nrows = H // ncols
    fig, fig_axes = plt.subplots(
        figsize=(4.5 * ncols, 4.5 * nrows),
        ncols=ncols,
        nrows=nrows,
        sharex=True,
        sharey=True,
    )
    fig.subplots_adjust(hspace=0.33, wspace=0.25)
    cbar_ax = fig.add_axes([0.92, 0.3, 0.03, 0.4])
    atom_fracs = get_atomic_fracs(data_loader, idx=idx)
    form = get_form(data_loader, idx=idx)
    atomic_numbers = get_atomic_numbers(data_loader, idx=idx).ravel().tolist()
    cons = CONSTANTS()
    idx_symbol_dict = cons.idx_symbol_dict
    atoms = [idx_symbol_dict[num] for num in atomic_numbers]

    if mask:
        atom_presence = np.array(atom_fracs > 0)
        mask = atom_presence * atom_presence.T

    plot_four = True
    label_abcd = ["a", "b", "c", "d"]
    if plot_four:
        for h in range(4):
            map_data = get_attention(attn_mat, idx=idx, layer=layer, head=h)
            n_el = 4
            plot_attention(
                map_data[:n_el, :n_el],
                # xlabel='fractional amount',
                # ylabel='atoms',
                xticklabels=atom_fracs.ravel()[:n_el],
                yticklabels=atoms[:n_el],
                mask=mask[:n_el, :n_el],
                ax=fig_axes.ravel()[h],
                cbar_ax=cbar_ax,
            )
            fig_axes.ravel()[h].set_title(label=f"layer {layer}, head {h}")
            fig_axes.ravel()[h].set_title(
                label=f'{label_abcd[h]}){40*" "}',
                fontdict={"fontweight": "bold"},
                y=1.05,
            )
        plt.savefig(
            "figures/Figure1_attention_plot_Al2O3.png", bbox_inches="tight", dpi=300
        )
        plt.show()
        # exit
    else:
        for h in range(H):
            map_data = get_attention(attn_mat, idx=idx, layer=layer, head=h)
            plot_attention(
                map_data,
                xlabel="fractional amount",
                ylabel="atoms",
                xticklabels=atom_fracs.ravel(),
                yticklabels=atoms,
                mask=mask,
                ax=fig_axes.ravel()[h],
                cbar_ax=cbar_ax,
            )
            fig_axes.ravel()[h].set_title(label=f"layer {layer}, head {h}")
        fig.suptitle(f"index: {idx}, formula: {form}")
        plt.savefig(
            "figures/Figure1_attention_plot_Al2O3.png", bbox_inches="tight", dpi=300
        )
        plt.show()
    return fig, fig_axes


# %%
formula_idx = [(i, out[2]) for i, out in enumerate(data_loader.dataset)]
find_formula = "Al2O3"
idx = [i for i, out in enumerate(data_loader.dataset) if out[2] == find_formula]
if len(idx) > 0:
    print(f"Found {find_formula} at index {idx}")

for i in range(len(idx)):
    for n in range(N):
        plot_all_heads(data_loader, attn_data, idx=idx[i], layer=n)
        break
    break



import warnings

import pandas as pd
import seaborn as sns

import matplotlib.pyplot as plt

from sklearn.model_selection import GridSearchCV

plt.rcParams.update({"font.size": 16})


# %%
class EstimatorSelectionHelper:
    def __init__(self, models, params, random_seed=42):
        self.models = models
        self.params = params
        self.keys = models.keys()
        self.grid_searches = {}
        if not set(models.keys()).issubset(set(params.keys())):
            missing_params = list(set(models.keys()) - set(params.keys()))
            raise ValueError(f"Missing estimator parameters: {missing_params}")

    def fit(
        self, X, y, cv=3, n_jobs=1, verbose=1, scoring=None, refit=False, random_seed=42
    ):
        if n_jobs == 1:
            warnings.warn(
                "n_jobs is currently set at 1, "
                "consider passing n_jobs=-1 or n_jobs=n_cores, "
                "where n_cores is the number of CPU cores you have",
                RuntimeWarning,
            )
        for key in self.keys:
            print(f"\nRunning GridSearchCV for {key}.")
            model = self.models[key]
            params = self.params[key]
            gs = GridSearchCV(
                model,
                params,
                cv=cv,
                n_jobs=n_jobs,
                verbose=verbose,
                scoring=scoring,
                refit=refit,
                return_train_score=True,
            )
            gs.fit(X, y)
            self.grid_searches[key] = gs
        self.scoring = scoring
        self.scoring_val2key = {scoring[k]: k for k in scoring}

    def plot_gridsearch(self, model_name, elem_prop, mat_prop, fig_dir, gs):
        test_metric = "mean_test_neg_MAE"

        # Get CV parameters and results
        dims = pd.DataFrame(gs.cv_results_["params"])
        # for RandomForest: convert max_depth = None to string 'None'
        dims.loc[dims["max_depth"].isnull(), "max_depth"] = "None"
        dims["score"] = gs.cv_results_[test_metric]
        col_names = dims.columns.tolist()
        dims = dims.pivot(col_names[0], col_names[1], col_names[2])
        dims_flipped = dims.iloc[::-1]

        # Plot and save gridsearch heatmap
        plt.figure(figsize=(6, 6))
        sns.heatmap(dims_flipped)
        plt.title(
            f"{test_metric}\n" f"{model_name} on {elem_prop} with {mat_prop}",
            fontsize=18,
        )
        plt.savefig(
            fig_dir + f"{model_name}_{elem_prop}_{mat_prop}.png",
            dpi=300,
            bbox_inches="tight",
        )
        plt.close("all")
        print(f"saved figure {model_name}_{elem_prop}_{mat_prop}.png")

    def score_summary(self, ep, mp, fig_dir, sort_by="mean_test_r2"):
        print("***************** gridsearch done *****************")
        scoring_keys = self.scoring.keys()

        col_prefixes = ["mean_", "std_"]
        col_midfixes = ["train_", "test_"]
        col_suffixes = scoring_keys
        columns1 = ["estimator"]
        columns2 = [
            p + m + s for s in col_suffixes for m in col_midfixes for p in col_prefixes
        ]
        all_columns = (
            columns1 + columns2 + ["params"] + ["mean_fit_time", "mean_score_time"]
        )

        df = pd.DataFrame(columns=all_columns)
        df_best_models = pd.DataFrame(columns=all_columns)

        # Plot gridsearch results for each model/param/elem_prop/mat_prop
        # combination, and save dataframe of best results
        for m in self.grid_searches:
            self.plot_gridsearch(m, ep, mp, fig_dir, self.grid_searches[m])

            df_model = pd.DataFrame(columns=all_columns)
            print(f"Parsing results for {m}")
            res = self.grid_searches[m].cv_results_
            params = res["params"]
            df_model["estimator"] = [m] * len(params)
            for col in df_model.columns:
                if col != "estimator":
                    df_model[col] = res[col]

            best_run_idx = df_model["mean_test_neg_MAE"].idxmax()
            best_run = df_model.loc[[best_run_idx], :]
            df_best_models = pd.concat([df_best_models, best_run], axis=0)
            df = pd.concat([df, df_model], axis=0)

        df["elem_prop"] = ep
        df["mat_prop"] = mp
        df_best_models["elem_prop"] = ep
        df_best_models["mat_prop"] = mp

        return df, df_best_models



import os
import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.cm as cm

from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import Normalize

from .composition import _element_composition
from scipy import stats

plt.rcParams.update({"font.size": 14})


# %%
def act_pred(
    y_act,
    y_pred,
    name="example",
    x_hist=True,
    y_hist=True,
    reg_line=True,
    save_dir=None,
):

    mec = "#2F4F4F"
    mfc = "#C0C0C0"

    plt.figure(1, figsize=(4, 4))
    left, width = 0.1, 0.65

    bottom, height = 0.1, 0.65
    bottom_h = left_h = left + width
    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom_h, width, 0.15]
    rect_histy = [left_h, bottom, 0.15, height]

    ax2 = plt.axes(rect_scatter)
    ax2.tick_params(direction="in", length=7, top=True, right=True)

    # add minor ticks
    minor_locator_x = AutoMinorLocator(2)
    minor_locator_y = AutoMinorLocator(2)
    ax2.get_xaxis().set_minor_locator(minor_locator_x)
    ax2.get_yaxis().set_minor_locator(minor_locator_y)
    plt.tick_params(which="minor", direction="in", length=4, right=True, top=True)

    # feel free to change the colors here.
    ax2.plot(
        y_act, y_pred, "o", mfc=mfc, alpha=0.5, label=None, mec=mec, mew=1.2, ms=5.2
    )

    ax2.plot(
        [-(10 ** 9), 10 ** 9], [-(10 ** 9), 10 ** 9], "k--", alpha=0.8, label="ideal"
    )

    ax2.set_ylabel("Predicted value (Units)")
    ax2.set_xlabel("Actual value (Units)")
    x_range = max(y_act) - min(y_act)
    ax2.set_xlim(max(y_act) - x_range * 1.05, min(y_act) + x_range * 1.05)

    ax2.set_ylim(max(y_act) - x_range * 1.05, min(y_act) + x_range * 1.05)

    ax1 = plt.axes(rect_histx)
    ax1_n, ax1_bins, ax1_patches = ax1.hist(
        y_act, bins=31, density=True, color=mfc, edgecolor=mec, alpha=0
    )
    ax1.set_xticks([])
    ax1.set_yticks([])
    ax1.set_xlim(ax2.get_xlim())
    ax1.axis("off")

    if x_hist:
        [p.set_alpha(1.0) for p in ax1_patches]

    ax3 = plt.axes(rect_histy)
    ax3_n, ax3_bins, ax3_patches = ax3.hist(
        y_pred,
        bins=31,
        density=True,
        color=mfc,
        edgecolor=mec,
        orientation="horizontal",
        alpha=0,
    )
    ax3.set_xticks([])
    ax3.set_yticks([])
    ax3.set_ylim(ax2.get_ylim())
    ax3.axis("off")

    if y_hist:
        [p.set_alpha(1.0) for p in ax3_patches]

    if reg_line:
        polyfit = np.polyfit(y_act, y_pred, deg=1)
        reg_ys = np.poly1d(polyfit)(np.unique(y_act))
        ax2.plot(np.unique(y_act), reg_ys, alpha=0.8, label="linear fit")

    ax2.legend(loc=2, framealpha=0.35, handlelength=1.5)
    plt.draw()

    if save_dir is not None:
        fig_name = f"{save_dir}/{name}_act_pred.png"
        os.makedirs(save_dir, exist_ok=True)
        plt.savefig(fig_name, bbox_inches="tight", dpi=300)
    plt.draw()
    plt.pause(0.001)
    plt.close()


def residual(y_act, y_pred, name="example", save_dir=None):

    mec = "#2F4F4F"
    mfc = "#C0C0C0"

    y_act = np.array(y_act)
    y_pred = np.array(y_pred)

    xmin = np.min([y_act]) * 0.9
    xmax = np.max([y_act]) / 0.9

    y_err = y_pred - y_act
    ymin = np.min([y_err]) * 0.9
    ymax = np.max([y_err]) / 0.9

    fig, ax = plt.subplots(figsize=(4, 4))

    ax.plot(y_act, y_err, "o", mec=mec, mfc=mfc, alpha=0.5, label=None, mew=1.2, ms=5.2)
    ax.plot([xmin, xmax], [0, 0], "k--", alpha=0.8, label="ideal")

    ax.set_ylabel("Residual error (Units)")
    ax.set_xlabel("Actual value (Units)")
    ax.legend(loc="lower right")

    minor_locator_x = AutoMinorLocator(2)
    minor_locator_y = AutoMinorLocator(2)
    ax.get_xaxis().set_minor_locator(minor_locator_x)
    ax.get_yaxis().set_minor_locator(minor_locator_y)

    ax.tick_params(right=True, top=True, direction="in", length=7)
    ax.tick_params(which="minor", right=True, top=True, direction="in", length=4)

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)

    if save_dir is not None:
        fig_name = f"{save_dir}/{name}_residual.png"
        os.makedirs(save_dir, exist_ok=True)
        plt.savefig(fig_name, bbox_inches="tight", dpi=300)
    plt.draw()
    plt.pause(0.001)
    plt.close()


def residual_hist(y_act, y_pred, name="example", save_dir=None):

    mec = "#2F4F4F"
    mfc = "#C0C0C0"

    fig, ax = plt.subplots(figsize=(4, 4))
    y_err = y_pred - y_act
    kde_act = stats.gaussian_kde(y_err)
    x_range = np.linspace(min(y_err), max(y_err), 1000)

    ax.hist(y_err, color=mfc, bins=35, alpha=1, edgecolor=mec, density=True)
    ax.plot(x_range, kde_act(x_range), "-", lw=1.2, color="k", label="kde")

    ax.set_xlabel("Residual error (Units)")
    plt.legend(loc=2, framealpha=0.35, handlelength=1.5)

    ax.tick_params(direction="in", length=7, top=True, right=True)

    minor_locator_x = AutoMinorLocator(2)
    minor_locator_y = AutoMinorLocator(2)
    ax.get_xaxis().set_minor_locator(minor_locator_x)
    ax.get_yaxis().set_minor_locator(minor_locator_y)
    plt.tick_params(which="minor", direction="in", length=4, right=True, top=True)

    if save_dir is not None:
        fig_name = f"{save_dir}/{name}_residual_hist.png"
        os.makedirs(save_dir, exist_ok=True)
        plt.savefig(fig_name, bbox_inches="tight", dpi=300)

    plt.draw()
    plt.pause(0.001)
    plt.close()


def loss_curve(x_data, train_err, val_err, name="example", save_dir=None):

    mec1 = "#2F4F4F"
    mfc1 = "#C0C0C0"
    mec2 = "maroon"
    mfc2 = "pink"

    fig, ax = plt.subplots(figsize=(4, 4))

    ax.plot(
        x_data,
        train_err,
        "-",
        color=mec1,
        marker="o",
        mec=mec1,
        mfc=mfc1,
        ms=4,
        alpha=0.5,
        label="train",
    )
    ax.plot(
        x_data,
        val_err,
        "--",
        color=mec2,
        marker="s",
        mec=mec2,
        mfc=mfc2,
        ms=4,
        alpha=0.5,
        label="validation",
    )

    max_val_err = max(val_err)
    ax.axhline(max_val_err, color="b", linestyle="--", alpha=0.3)

    ax.set_xlabel("Number of training epochs")
    ax.set_ylabel("Loss (Units)")
    ax.set_ylim(0, 2 * np.mean(val_err))

    ax.legend(loc=1, framealpha=0.35, handlelength=1.5)

    minor_locator_x = AutoMinorLocator(2)
    minor_locator_y = AutoMinorLocator(2)
    ax.get_xaxis().set_minor_locator(minor_locator_x)
    ax.get_yaxis().set_minor_locator(minor_locator_y)

    ax.tick_params(right=True, top=True, direction="in", length=7)
    ax.tick_params(which="minor", right=True, top=True, direction="in", length=4)

    if save_dir is not None:
        fig_name = f"{save_dir}/{name}_loss_curve.png"
        os.makedirs(save_dir, exist_ok=True)
        plt.savefig(fig_name, bbox_inches="tight", dpi=300)
    plt.draw()
    plt.pause(0.001)
    plt.close()


def element_prevalence(
    formulae, name="example", save_dir=None, log_scale=False, ptable_fig=True
):
    ptable = pd.read_csv("ML_figures/element_properties/ptable.csv")
    ptable.index = ptable["symbol"].values
    elem_tracker = ptable["count"]
    n_row = ptable["row"].max()
    n_column = ptable["column"].max()

    for formula in formulae:
        formula_dict = _element_composition(formula)
        elem_count = pd.Series(formula_dict, name="count")
        elem_tracker = elem_tracker.add(elem_count, fill_value=0)

    if ptable_fig:
        fig, ax = plt.subplots(figsize=(n_column, n_row))
        rows = ptable["row"]
        columns = ptable["column"]
        symbols = ptable["symbol"]
        rw = 0.9  # rectangle width (rw)
        rh = rw  # rectangle height (rh)
        for row, column, symbol in zip(rows, columns, symbols):
            row = ptable["row"].max() - row
            cmap = cm.YlGn
            count_min = elem_tracker.min()
            count_max = elem_tracker.max()
            norm = Normalize(vmin=count_min, vmax=count_max)
            count = elem_tracker[symbol]
            if log_scale:
                norm = Normalize(vmin=np.log(1), vmax=np.log(count_max))
                if count != 0:
                    count = np.log(count)
            color = cmap(norm(count))
            if count == 0:
                color = "silver"
            if row < 3:
                row += 0.5
            rect = patches.Rectangle(
                (column, row),
                rw,
                rh,
                linewidth=1.5,
                edgecolor="gray",
                facecolor=color,
                alpha=1,
            )

            plt.text(
                column + rw / 2,
                row + rw / 2,
                symbol,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=20,
                fontweight="semibold",
                color="k",
            )

            ax.add_patch(rect)

        granularity = 20
        for i in range(granularity):
            value = int(round((i) * count_max / (granularity - 1)))
            if log_scale:
                if value != 0:
                    value = np.log(value)
            color = cmap(norm(value))
            if value == 0:
                color = "silver"
            length = 9
            x_offset = 3.5
            y_offset = 7.8
            x_loc = i / (granularity) * length + x_offset
            width = length / granularity
            height = 0.35
            rect = patches.Rectangle(
                (x_loc, y_offset),
                width,
                height,
                linewidth=1.5,
                edgecolor="gray",
                facecolor=color,
                alpha=1,
            )

            if i in [0, 4, 9, 14, 19]:
                text = f"{value:0.0f}"
                if log_scale:
                    text = f"{np.exp(value):0.1e}".replace("+", "")
                plt.text(
                    x_loc + width / 2,
                    y_offset - 0.4,
                    text,
                    horizontalalignment="center",
                    verticalalignment="center",
                    fontweight="semibold",
                    fontsize=20,
                    color="k",
                )

            ax.add_patch(rect)

        plt.text(
            x_offset + length / 2,
            y_offset + 0.7,
            "log(Element Count)" if log_scale else "Element Count",
            horizontalalignment="center",
            verticalalignment="center",
            fontweight="semibold",
            fontsize=20,
            color="k",
        )

        ax.set_ylim(-0.15, n_row + 0.1)
        ax.set_xlim(0.85, n_column + 1.1)

        # fig.patch.set_visible(False)
        ax.axis("off")

        if save_dir is not None:
            fig_name = f"{save_dir}/{name}_ptable.png"
            os.makedirs(save_dir, exist_ok=True)
            plt.savefig(fig_name, bbox_inches="tight", dpi=300)
        plt.draw()
        plt.pause(0.001)
        plt.close()

    if not ptable_fig:
        fig, ax = plt.subplots(figsize=(15, 6))
        non_zero = elem_tracker[elem_tracker != 0].sort_values(ascending=False)
        if log_scale:
            non_zero = np.log(non_zero)
        non_zero.plot.bar(width=0.7, edgecolor="k")

        minor_locator_y = AutoMinorLocator(2)
        ax.get_yaxis().set_minor_locator(minor_locator_y)

        ax.set_ylabel("Element Count")
        if log_scale:
            ax.set_ylabel("log(Element Count)")

        ax.tick_params(right=True, top=True, direction="in", length=7)
        ax.tick_params(which="minor", right=True, top=True, direction="in", length=4)

        if save_dir is not None:
            fig_name = f"{save_dir}/{name}_elem_hist.png"
            os.makedirs(save_dir, exist_ok=True)
            plt.savefig(fig_name, bbox_inches="tight", dpi=300)
        plt.draw()
        plt.pause(0.001)
        plt.close()


# %%
if __name__ == "__main__":
    # read in example act vs. pred data
    df_act_pred = pd.read_csv("example_data/act_pred.csv")
    y_act, y_pred = df_act_pred.iloc[:, 1], df_act_pred.iloc[:, 2]

    act_pred(y_act, y_pred, reg_line=True, save_dir="example_figures")

    act_pred(
        y_act,
        y_pred,
        name="example_no_hist",
        x_hist=False,
        y_hist=False,
        reg_line=True,
        save_dir="example_figures",
    )

    residual(y_act, y_pred, save_dir="example_figures")

    residual_hist(y_act, y_pred, save_dir="example_figures")

    # read in loss curve data
    df_lc = pd.read_csv("example_data/training_progress.csv")
    epoch = df_lc["epoch"]
    train_err, val_err = df_lc["mae_train"], df_lc["mae_val"]

    loss_curve(epoch, train_err, val_err, save_dir="example_figures")

    # element prevalence
    formula = df_act_pred.iloc[:, 0]

    element_prevalence(formula, save_dir="example_figures", log_scale=False)
    element_prevalence(
        formula, save_dir="example_figures", name="example_log", log_scale=True
    )

    plt.rcParams.update({"font.size": 12})
    element_prevalence(
        formula, save_dir="example_figures", ptable_fig=False, log_scale=False
    )
    element_prevalence(
        formula,
        save_dir="example_figures",
        name="example_log",
        ptable_fig=False,
        log_scale=True,
    )



import os

from time import time

from crabnet.utils.utils import get_cbfv

from crabnet.utils.estimatorselectionhelper import EstimatorSelectionHelper


# %%
def modelselectionhelper(
    models,
    params,
    elem_props,
    mat_props_dir,
    mat_props,
    metrics_dir,
    fig_dir,
    scoring=None,
    n_jobs=1,
    cv=3,
    refit="neg_MAE",
    verbose=False,
    random_seed=42,
):
    if scoring is None:
        scoring = {"neg_MAE": "neg_mean_absolute_error"}
    for ep in elem_props:
        print("\n++++++++++++++++++++++++++++++++++++++++++++++++")
        print(f"currently using element property: {ep}")
        print("++++++++++++++++++++++++++++++++++++++++++++++++")
        ti_ep = time()

        for mp in mat_props:
            print(f"fitting {mp} with {ep} using {cv}-fold CV")
            ti_mp = time()

            trainpath = os.path.join(mat_props_dir, mp, "train.csv")
            valpath = os.path.join(mat_props_dir, mp, "val.csv")

            if not os.path.exists(trainpath) or not os.path.exists(valpath):
                trainpath = os.path.join(mat_props_dir, mp, "train0.csv")
                valpath = os.path.join(mat_props_dir, mp, "val0.csv")

            X, y, form, skipped = get_cbfv(trainpath, elem_prop=ep)
            X_val, y_val, form_val, skipped_val = get_cbfv(valpath, elem_prop=ep)

            # Sample the dataset for faster gridsearch
            n_samples = 2000
            if X.shape[0] > n_samples:
                print(
                    f"Sampling training data to {n_samples} samples "
                    f"to speed up initial gridsearch"
                )
                X = X.sample(n=n_samples)
                y = y.loc[X.index]
                form = form.loc[X.index]

            helper1 = EstimatorSelectionHelper(models, params)
            helper1.fit(
                X,
                y,
                scoring=scoring,
                n_jobs=n_jobs,
                cv=cv,
                refit=refit,
                verbose=verbose,
                random_seed=random_seed,
            )

            output = helper1.score_summary(ep, mp, fig_dir, sort_by="mean_test_r2")
            score_summary, best_models = output

            print("\n************************************************")
            print(f"finished {mp} with {ep}")
            print("saving score summary and best models files")
            print("************************************************")

            outpath_all = os.path.join(metrics_dir, f"{ep}_{mp}.csv")
            score_summary.to_csv(outpath_all, index=False)

            outpath_best = os.path.join(metrics_dir, f"best_{ep}_{mp}.csv")
            best_models.to_csv(outpath_best, index=False)

            dt_mp = time() - ti_mp
            print(f"time elapsed for {mp} with {ep}: {dt_mp:0.4f} s")

        dt_ep = time() - ti_ep
        print(
            f"time elapsed for all material properties " f"using {ep}: {dt_ep:0.4f} s"
        )

    return helper1



import numpy as np

import torch

from crabnet.kingcrab import CrabNet
from crabnet.model import Model
from crabnet.utils.get_compute_device import get_compute_device
from crabnet.utils.composition import _element_composition

import matplotlib.pyplot as plt

from collections import defaultdict

compute_device = get_compute_device(prefer_last=False)
data_type_torch = torch.float32


# %%
# fmt: off
all_symbols = ['None', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na',
               'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc',
               'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
               'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb',
               'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb',
               'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
               'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
               'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl',
               'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
               'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
               'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg',
               'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
# fmt: on

color = [
    "#e41a1c",
    "#377eb8",
    "#4daf4a",
    "#984ea3",
    "#ff7f00",
    "#ffff33",
    "#a65628",
    "#f781bf",
]

classification_list = []

num = ""
mat_prop = "aflow__Egap"

# Get the TorchedCrabNet architecture loaded
model = Model(CrabNet().to(compute_device), model_name=f"{mat_prop}")
if True:
    model.load_network(f"{mat_prop}{num}.pth")
    model.model_name = f"{mat_prop}{num}"

if mat_prop in classification_list:
    model.classification = True

mat_prop = "aflow__Egap"
test_data = rf"data\benchmark_data\{mat_prop}\train.csv"
# test_data = rf'data\matbench_cv\{mat_prop}\train{num}.csv'

model.load_data(test_data, batch_size=2 ** 0)  # data is reloaded to model.data_loader

len_dataset = len(model.data_loader.dataset)
n_atoms = int(len(model.data_loader.dataset[0][0]) / 2)
act = np.zeros(len_dataset)
pred = np.zeros(len_dataset)
uncert = np.zeros(len_dataset)
formulae = np.empty(len_dataset, dtype=list)
atoms = np.empty((len_dataset, n_atoms))
fractions = np.empty((len_dataset, n_atoms))
model.model.eval()
model.model.avg = False

simple_tracker = {i: [] for i in range(119)}
variance_tracker = {i: [] for i in range(119)}
element_tracker = {i: [] for i in range(119)}
composition_tracker = {}

alls_dict = defaultdict(list)
binaries_dict = defaultdict(list)
for data in model.data_loader.dataset:
    keys = _element_composition(data[2]).keys()
    if len(keys) == 2:
        binaries_dict[tuple(sorted(keys))].append(data)
    else:
        alls_dict[tuple(sorted(keys))].append(data)


# %%
max_len = 1

for key, systems in binaries_dict.items():
    if "Si" not in key:
        continue
    if "O" not in key:
        continue
    print(key)
    if len(systems) < max_len:
        continue
    og_points = []
    with torch.no_grad():
        for data in systems:
            X, y, formula = data
            print(formula)
            src, frac = X.unsqueeze(0).squeeze(-1).chunk(2, dim=1)
            src = src.to(compute_device, dtype=torch.long, non_blocking=True)
            frac = frac.to(compute_device, dtype=data_type_torch, non_blocking=True)
            y = y.to(compute_device, dtype=data_type_torch, non_blocking=True)
            output = model.model.forward(src, frac)
            mask = (src == 0).unsqueeze(-1).repeat(1, 1, 3)
            mask = (src == 0).unsqueeze(-1).repeat(1, 1, 1)
            output = output.masked_fill(mask, 0)
            prob = output[:, :, -1:]
            output = output[:, :, :2]
            probability = torch.ones_like(output)
            probability[:, :, :1] = torch.sigmoid(prob)
            output = output * probability

            prediction, uncertainty = output.chunk(2, dim=-1)
            uncertainty = torch.exp(uncertainty) * model.scaler.std
            prediction = model.scaler.unscale(prediction)
            prediction = prediction * ~mask
            uncertainty = uncertainty * ~mask
            slist = src.cpu().numpy().ravel().tolist()
            flist = frac.cpu().numpy().ravel().tolist()
            sf_dict = {src: frac for src, frac in zip(slist, flist)}
            elem_frac = sf_dict[sorted(slist)[-1]]
            og_points.append([float(elem_frac), float(y), formula])

        frac_tracker = []
        pred_tracker = []
        uncertainty_tracker = []
        elem1_tracker = []
        elem2_tracker = []
        res = 50
        X, y, formula = data
        X_data = [(src, frac) for src, frac in zip(slist, flist)]
        X_data = sorted(X_data, reverse=True)
        slist = [x[0] for x in X_data]
        flist = [x[1] for x in X_data]
        X = torch.tensor(slist + flist).unsqueeze(1)

        first = True
        for i in range(0, res - 1):
            src, frac = X.unsqueeze(0).squeeze(-1).chunk(2, dim=1)
            frac = frac.clone()
            src = src.clone()

            frac[:, 0:1] = 1 / res + (i / res)
            frac[:, 1:2] = 1 - frac[:, 0:1]
            src = src.to(compute_device, dtype=torch.long, non_blocking=True)
            frac = frac.to(compute_device, dtype=data_type_torch, non_blocking=True)
            y = y.to(compute_device, dtype=data_type_torch, non_blocking=True)

            output = model.model.forward(src, frac)
            mask = (src == 0).unsqueeze(-1).repeat(1, 1, 1)
            output = output.masked_fill(mask, 0)
            prob = output[:, :, -1:]
            output = output[:, :, :2]
            probability = torch.ones_like(output)
            probability[:, :, :1] = torch.sigmoid(prob)
            output = output * probability

            prediction, uncertainty = output.chunk(2, dim=-1)
            prediction = model.scaler.unscale(prediction)
            prediction = prediction * ~mask
            uncertainty = uncertainty * ~mask
            uncertainty_mean = (uncertainty * ~mask).sum() / (~mask).sum()
            uncertainty = torch.exp(uncertainty_mean) * model.scaler.std

            min_val = prediction[:, :, :][prediction != 0].min().cpu().detach().numpy()
            max_val = prediction[:, :, :][prediction != 0].max().cpu().detach().numpy()
            range_val = max_val - min_val
            pred_elem1 = prediction[:, 0, :].cpu().detach().numpy().ravel()
            pred_elem2 = prediction[:, 1, :].cpu().detach().numpy().ravel()
            act_val = y.detach().cpu().numpy().ravel()
            uncertainty_tracker.append(uncertainty.cpu().detach().numpy())
            frac_tracker.append(frac[0, :1].cpu().detach().numpy())
            elem1_tracker.append(pred_elem1)
            elem2_tracker.append(pred_elem2)
            pred_tracker.append(
                ((prediction * ~mask).sum() / (~mask).sum()).cpu().detach().numpy()
            )
            elem1_name = all_symbols[int(src[0, 0].cpu().detach().numpy())]
            elem2_name = all_symbols[int(src[0, 1].cpu().detach().numpy())]

        frac_tracker = np.array(frac_tracker).ravel()
        pred_tracker = np.array(pred_tracker).ravel()
        uncertainty_tracker = np.array(uncertainty_tracker).ravel()
        plt.figure(figsize=(7, 5))
        plt.fill_between(
            frac_tracker,
            pred_tracker - uncertainty_tracker,
            pred_tracker + uncertainty_tracker,
            alpha=0.7,
            color="silver",
        )
        plt.plot(
            frac_tracker,
            elem1_tracker,
            "-.",
            label=f"{elem1_name} contribution",
            color=color[0],
            linewidth=2,
        )
        plt.plot(
            frac_tracker,
            elem2_tracker,
            "--",
            label=f"{elem2_name} contribution",
            color=color[1],
            linewidth=2,
        )
        plt.plot(
            frac_tracker,
            pred_tracker,
            "-",
            color="gray",
            label="Prediction",
            linewidth=2,
        )
        plt.xlim(0, 1)
        plt.ylabel("$Si_xO_{1-x}$ Band Gap (eV)")
        plt.xlabel("$x$")

        plt.legend(
            ncol=1, handlelength=2, labelspacing=0.08, columnspacing=1, framealpha=0.3
        )
        plt.savefig(
            f"figures/Figure4_binary_plot_{elem1_name}-{elem2_name}.png",
            bbox_inches="tight",
            dpi=300,
        )
        plt.show()



from os.path import join, dirname

import numpy as np
import pandas as pd

import torch
from torch import nn


# %%
RNG_SEED = 42
torch.manual_seed(RNG_SEED)
np.random.seed(RNG_SEED)
data_type_torch = torch.float32


# %%
class ResidualNetwork(nn.Module):
    """
    Feed forward Residual Neural Network as seen in Roost.
    https://doi.org/10.1038/s41467-020-19964-7
    """

    def __init__(self, input_dim, output_dim, hidden_layer_dims):
        """
        Inputs
        ----------
        input_dim: int
        output_dim: int
        hidden_layer_dims: list(int)
        """
        super(ResidualNetwork, self).__init__()
        dims = [input_dim] + hidden_layer_dims
        self.fcs = nn.ModuleList(
            [nn.Linear(dims[i], dims[i + 1]) for i in range(len(dims) - 1)]
        )
        self.res_fcs = nn.ModuleList(
            [
                nn.Linear(dims[i], dims[i + 1], bias=False)
                if (dims[i] != dims[i + 1])
                else nn.Identity()
                for i in range(len(dims) - 1)
            ]
        )
        self.acts = nn.ModuleList([nn.LeakyReLU() for _ in range(len(dims) - 1)])
        self.fc_out = nn.Linear(dims[-1], output_dim)

    def forward(self, fea):
        for fc, res_fc, act in zip(self.fcs, self.res_fcs, self.acts):
            fea = act(fc(fea)) + res_fc(fea)
        return self.fc_out(fea)

    def __repr__(self):
        return f"{self.__class__.__name__}"


class Embedder(nn.Module):
    def __init__(self, d_model, compute_device=None):
        super().__init__()
        self.d_model = d_model
        self.compute_device = compute_device

        elem_dir = join(dirname(__file__), "data", "element_properties")
        # # Choose what element information the model receives
        mat2vec = join(elem_dir, "mat2vec.csv")  # element embedding
        # mat2vec = f'{elem_dir}/onehot.csv'  # onehot encoding (atomic number)
        # mat2vec = f'{elem_dir}/random_200.csv'  # random vec for elements

        cbfv = pd.read_csv(mat2vec, index_col=0).values
        feat_size = cbfv.shape[-1]
        self.fc_mat2vec = nn.Linear(feat_size, d_model).to(self.compute_device)
        zeros = np.zeros((1, feat_size))
        cat_array = np.concatenate([zeros, cbfv])
        cat_array = torch.as_tensor(cat_array, dtype=data_type_torch)
        self.cbfv = nn.Embedding.from_pretrained(cat_array).to(
            self.compute_device, dtype=data_type_torch
        )

    def forward(self, src):
        mat2vec_emb = self.cbfv(src)
        x_emb = self.fc_mat2vec(mat2vec_emb)
        return x_emb


# %%
class FractionalEncoder(nn.Module):
    """
    Encoding element fractional amount using a "fractional encoding" inspired
    by the positional encoder discussed by Vaswani.
    https://arxiv.org/abs/1706.03762
    """

    def __init__(self, d_model, resolution=100, log10=False, compute_device=None):
        super().__init__()
        self.d_model = d_model // 2
        self.resolution = resolution
        self.log10 = log10
        self.compute_device = compute_device

        x = torch.linspace(
            0, self.resolution - 1, self.resolution, requires_grad=False
        ).view(self.resolution, 1)
        fraction = (
            torch.linspace(0, self.d_model - 1, self.d_model, requires_grad=False)
            .view(1, self.d_model)
            .repeat(self.resolution, 1)
        )

        pe = torch.zeros(self.resolution, self.d_model)
        pe[:, 0::2] = torch.sin(x / torch.pow(50, 2 * fraction[:, 0::2] / self.d_model))
        pe[:, 1::2] = torch.cos(x / torch.pow(50, 2 * fraction[:, 1::2] / self.d_model))
        pe = self.register_buffer("pe", pe)

    def forward(self, x):
        x = x.clone()
        if self.log10:
            x = 0.0025 * (torch.log2(x)) ** 2
            x[x > 1] = 1
            # x = 1 - x  # for sinusoidal encoding at x=0
        x[x < 1 / self.resolution] = 1 / self.resolution
        frac_idx = torch.round(x * (self.resolution)).to(dtype=torch.long) - 1
        out = self.pe[frac_idx]

        return out


# %%
class Encoder(nn.Module):
    def __init__(self, d_model, N, heads, frac=False, attn=True, compute_device=None):
        super().__init__()
        self.d_model = d_model
        self.N = N
        self.heads = heads
        self.fractional = frac
        self.attention = attn
        self.compute_device = compute_device
        self.embed = Embedder(d_model=self.d_model, compute_device=self.compute_device)
        self.pe = FractionalEncoder(self.d_model, resolution=5000, log10=False)
        self.ple = FractionalEncoder(self.d_model, resolution=5000, log10=True)

        self.emb_scaler = nn.parameter.Parameter(torch.tensor([1.0]))
        self.pos_scaler = nn.parameter.Parameter(torch.tensor([1.0]))
        self.pos_scaler_log = nn.parameter.Parameter(torch.tensor([1.0]))

        if self.attention:
            encoder_layer = nn.TransformerEncoderLayer(
                self.d_model, nhead=self.heads, dim_feedforward=2048, dropout=0.1
            )
            self.transformer_encoder = nn.TransformerEncoder(
                encoder_layer, num_layers=self.N
            )

    def forward(self, src, frac):
        x = self.embed(src) * 2 ** self.emb_scaler
        mask = frac.unsqueeze(dim=-1)
        mask = torch.matmul(mask, mask.transpose(-2, -1))
        mask[mask != 0] = 1
        src_mask = mask[:, 0] != 1

        pe = torch.zeros_like(x)
        ple = torch.zeros_like(x)
        pe_scaler = 2 ** (1 - self.pos_scaler) ** 2
        ple_scaler = 2 ** (1 - self.pos_scaler_log) ** 2
        pe[:, :, : self.d_model // 2] = self.pe(frac) * pe_scaler
        ple[:, :, self.d_model // 2 :] = self.ple(frac) * ple_scaler

        if self.attention:
            x_src = x + pe + ple
            x_src = x_src.transpose(0, 1)
            x = self.transformer_encoder(x_src, src_key_padding_mask=src_mask)
            x = x.transpose(0, 1)

        if self.fractional:
            x = x * frac.unsqueeze(2).repeat(1, 1, self.d_model)

        hmask = mask[:, :, 0:1].repeat(1, 1, self.d_model)
        if mask is not None:
            x = x.masked_fill(hmask == 0, 0)

        return x


# %%
class CrabNet(nn.Module):
    def __init__(self, out_dims=3, d_model=512, N=3, heads=4, compute_device=None):
        super().__init__()
        self.avg = True
        self.out_dims = out_dims
        self.d_model = d_model
        self.N = N
        self.heads = heads
        self.compute_device = compute_device
        self.encoder = Encoder(
            d_model=self.d_model,
            N=self.N,
            heads=self.heads,
            compute_device=self.compute_device,
        )
        self.out_hidden = [1024, 512, 256, 128]
        self.output_nn = ResidualNetwork(self.d_model, self.out_dims, self.out_hidden)

    def forward(self, src, frac):
        output = self.encoder(src, frac)

        # average the "element contribution" at the end
        # mask so you only average "elements"
        mask = (src == 0).unsqueeze(-1).repeat(1, 1, self.out_dims)
        output = self.output_nn(output)  # simple linear
        if self.avg:
            output = output.masked_fill(mask, 0)
            output = output.sum(dim=1) / (~mask).sum(dim=1)
            output, logits = output.chunk(2, dim=-1)
            probability = torch.ones_like(output)
            probability[:, : logits.shape[-1]] = torch.sigmoid(logits)
            output = output * probability

        return output


# %%
if __name__ == "__main__":
    model = CrabNet()



import numpy as np
import pandas as pd
import os

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.colors import Normalize
import matplotlib.patheffects as path_effects
import seaborn as sns

from crabnet.kingcrab import CrabNet
from crabnet.model import Model
from crabnet.utils.get_compute_device import get_compute_device

import torch

from crabnet.utils.utils import CONSTANTS

compute_device = get_compute_device()
cons = CONSTANTS()


# %%
mat_prop = "aflow__Egap"
torchnet_params = {"d_model": 512, "N": 3, "heads": 4}
symbol_idx_dict = {val: key for key, val in cons.idx_symbol_dict.items()}

elem = "Si"
# elem = 23
if type(elem) == int:
    elem_sym = cons.idx_symbol_dict[elem]
    elem_Z = elem
elif type(elem) == str:
    elem_Z = symbol_idx_dict[elem]
    elem_sym = elem

train_data1 = rf"data\benchmark_data\{mat_prop}\train.csv"
datas = [train_data1]
in_fracs = ["Average Attention (AFLOW Egap)"]


class SaveOutput:
    def __init__(self):
        self.outputs = []

    def __call__(self, module, module_in, module_out):
        module_out = [out.detach().cpu() for out in module_out]
        self.outputs.append(module_out)

    def clear(self):
        self.outputs = []


save_output = SaveOutput()

for data, in_frac in zip(datas, in_fracs):

    # Create a model
    model = Model(
        CrabNet(**torchnet_params, compute_device=compute_device).to(compute_device)
    )
    model.load_network(f"{mat_prop}.pth")
    hook_handles = []

    # Insert forward hooks into model
    for layer in model.model.modules():
        if isinstance(layer, torch.nn.modules.activation.MultiheadAttention):
            # print('isinstance')
            handle = layer.register_forward_hook(save_output)
            hook_handles.append(handle)

    model.load_data(data)  # data is reloaded to model.data_loader

    save_output.clear()
    output = model.predict(model.data_loader)

    elem_pred = [
        (i, out, output[1][i]) for i, out in enumerate(output[2]) if elem_sym in out
    ]
    df_elem = pd.DataFrame(elem_pred, columns=["idx", "formula", "prediction"])

    mod_out = save_output.outputs

    n_mats = len(mod_out)  # number of output matrices from hook
    bsz = model.data_loader.batch_size  # batch size from data loader
    B = len(model.data_loader)  # total number of batches from data loader
    H = model.model.heads  # number of heads
    N = model.model.N  # number of layers
    n_data = len(model.data_loader.dataset)
    n_elements = model.n_elements

    assert n_mats == N * B, "something is wrong with the matrices"

    attn_data = torch.zeros(size=(n_data, N, H, n_elements, n_elements))
    for layer in range(N):
        sliceN = [
            save_output.outputs[i][1].unsqueeze(1) for i in range(layer, n_mats, N)
        ]
        sliceN = torch.cat(sliceN, dim=0)
        attn_data[:, layer : layer + 1, :, :, :] = sliceN

    save_output.clear()  # free up CPU RAM after getting attn info
    attn_data = attn_data.detach().cpu().numpy()
    data_loader = model.data_loader

    def get_datum(data_loader, idx=0):
        datum = data_loader.dataset[idx]
        return datum

    def get_x(data_loader, idx=0):
        x = get_datum(data_loader, idx=idx)[0]
        return x

    def get_atomic_numbers(data_loader, idx=0):
        nums = get_x(data_loader, idx=idx).chunk(2)[0].detach().cpu().numpy()
        nums = nums.astype(int)
        return nums

    def get_atomic_fracs(data_loader, idx=0):
        nums = get_x(data_loader, idx=idx).chunk(2)[1].detach().cpu().numpy()
        return nums

    def get_target(data_loader, idx=0):
        target = get_datum(data_loader, idx=idx)[1].detach().cpu().numpy()
        return target

    def get_form(data_loader, idx=0):
        form = get_datum(data_loader, idx=idx)[2]
        return form

    def get_attention(attn_mat, idx=0, layer=0, head=0):
        """
        Get one slice of the attention map.

        Parameters
        ----------
        attn_mat : Tensor
            attn_mat is numpy array in the shape of [S, N, H, d, d], where
            S is the total number of data samples,
            N is the layer number in the attention mechanism,
            H is the head number in the attention mechanism, and
            d is the attention dimension in each head.
        idx : int, optional
            Index of the input material. The default is 0.
        layer : int, optional
            Layer number in the attention mechanism. The default is 0.
        head : int, optional
            Head number in the attention mechanism. The default is 0.

        Returns
        -------
        attn : Tensor

        """
        attn_mat = attn_mat
        assert len(attn_mat.shape) == 5, "input attn_map is of the wrong shape"
        if head == "average":
            attn = attn_mat[idx, layer, :, :, :]
            attn = np.mean(attn, axis=0)
        elif isinstance(head, int):
            attn = attn_mat[idx, layer, head, :, :]
        return attn

    attn_mat = attn_data.copy()

    data_loader = model.data_loader

    idx = 1
    layer = 0

    other_dict = {i: [] for i in range(1, 119)}

    option = [0, 1, 2, 3, "average"]
    option_texts = ["a)", "b)", "c)", "d)", "average"]

    idx_plot = 0
    head_option = option[idx_plot]
    option_text = option_texts[idx_plot]

    for idx in range(len(data_loader.dataset)):
        map_data = get_attention(attn_mat, idx=idx, layer=layer, head=head_option)
        atom_fracs = get_atomic_fracs(data_loader, idx=idx)
        form = get_form(data_loader, idx=idx)
        atomic_numbers = get_atomic_numbers(data_loader, idx=idx).ravel().tolist()
        idx_symbol_dict = cons.idx_symbol_dict
        atoms = [idx_symbol_dict[num] for num in atomic_numbers]
        atom_presence = np.array(atom_fracs > 0)
        mask = atom_presence * atom_presence.T
        map_data = map_data * mask
        if elem_Z in atomic_numbers:
            row = atomic_numbers.index(elem_Z)
            for atomic_number in atomic_numbers:
                if atomic_number == 0:
                    continue
                col = atomic_numbers.index(atomic_number)
                # get the raw attention value
                other_dict[atomic_number].append(map_data[row, col])

    # fmt: off
    all_symbols = ['None', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na',
                   'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc',
                   'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
                   'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb',
                   'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb',
                   'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
                   'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
                   'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl',
                   'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
                   'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
                   'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg',
                   'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
    # fmt: on

    property_tracker = {
        all_symbols[key]: np.array(val).mean()
        for key, val in other_dict.items()
        if len(val) != 0
    }

    def plot(mat_prop, property_tracker):
        ptable = pd.read_csv("data/element_properties/ptable.csv")
        ptable.index = ptable["symbol"].values
        elem_tracker = ptable["count"]
        n_row = ptable["row"].max()
        n_column = ptable["column"].max()

        elem_tracker = elem_tracker + pd.Series(property_tracker)

        # log_scale = True
        log_scale = False

        fig, ax = plt.subplots(figsize=(n_column, n_row))
        rows = ptable["row"]
        columns = ptable["column"]
        symbols = ptable["symbol"]
        rw = 0.9  # rectangle width (rw)
        rh = rw  # rectangle height (rh)
        for row, column, symbol in zip(rows, columns, symbols):
            row = ptable["row"].max() - row
            cmap = sns.cm.rocket_r
            count_min = elem_tracker.min()
            count_max = elem_tracker.max()
            count_min = 0
            count_max = 1
            norm = Normalize(vmin=count_min, vmax=count_max)
            count = elem_tracker[symbol]
            if log_scale:
                norm = Normalize(vmin=np.log(1), vmax=np.log(count_max))
                if count != 0:
                    count = np.log(count)
            color = cmap(norm(count))
            if np.isnan(count):
                color = "silver"
            if row < 3:
                row += 0.5
            # element box
            rect = patches.Rectangle(
                (column, row),
                rw,
                rh,
                linewidth=1.5,
                edgecolor="gray",
                facecolor=color,
                alpha=1,
            )
            # plot element text
            text = plt.text(
                column + rw / 2,
                row + rw / 2,
                symbol,
                horizontalalignment="center",
                verticalalignment="center",
                fontsize=22,
                fontweight="semibold",
                color="white",
            )

            text.set_path_effects(
                [
                    path_effects.Stroke(linewidth=3, foreground="#030303"),
                    path_effects.Normal(),
                ]
            )

            ax.add_patch(rect)

        granularity = 20
        for i in range(granularity):
            value = (1 - i / (granularity - 1)) * count_min + (
                i / (granularity - 1)
            ) * count_max
            if log_scale:
                if value != 0:
                    value = np.log(value)
            color = cmap(norm(value))
            length = 9
            x_offset = 3.5
            y_offset = 7.8
            x_loc = i / (granularity) * length + x_offset
            width = length / granularity
            height = 0.35
            rect = patches.Rectangle(
                (x_loc, y_offset),
                width,
                height,
                linewidth=1.5,
                edgecolor="gray",
                facecolor=color,
                alpha=1,
            )

            if i in [0, 4, 9, 14, 19]:
                text = f"{value:0.2f}"
                if log_scale:
                    text = f"{np.exp(value):0.1e}".replace("+", "")
                plt.text(
                    x_loc + width / 2,
                    y_offset - 0.4,
                    text,
                    horizontalalignment="center",
                    verticalalignment="center",
                    fontweight="semibold",
                    fontsize=20,
                    color="k",
                )

            ax.add_patch(rect)

        legend_title = f"{elem_sym}, {in_frac}"
        plt.text(
            x_offset + length / 2,
            y_offset + 0.7,
            f"log({legend_title})" if log_scale else legend_title,
            horizontalalignment="center",
            verticalalignment="center",
            fontweight="semibold",
            fontsize=20,
            color="k",
        )
        # add annotation for subfigure numbering
        plt.text(
            0.55,
            n_row + 0.1,
            option_text,
            fontweight="semibold",
            fontsize=38,
            color="k",
        )
        ax.set_ylim(-0.15, n_row + 0.1)
        ax.set_xlim(0.85, n_column + 1.1)

        # fig.patch.set_visible(False)
        ax.axis("off")

        plt.draw()
        save_dir = "figures/"
        if save_dir is not None:
            fig_name = f"{save_dir}/Figure2_{mat_prop}_ptable_{head_option}.png"
            os.makedirs(save_dir, exist_ok=True)
            plt.savefig(fig_name, bbox_inches="tight", dpi=300)

        plt.pause(0.001)
        plt.close()

    plot(mat_prop, property_tracker)



from logging import warn
import os
from time import time
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# retrieve static file from package: https://stackoverflow.com/a/20885799/13697228
from importlib.resources import open_text

from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_absolute_error, roc_auc_score

import torch
from torch.optim.lr_scheduler import CyclicLR

from .utils.get_compute_device import get_compute_device

from .utils.utils import (
    Lamb,
    Lookahead,
    RobustL1,
    BCEWithLogitsLoss,
    EDM_CsvLoader,
    Scaler,
    DummyScaler,
    count_parameters,
)
from crabnet.utils.optim import SWA

from crabnet.kingcrab import CrabNet

# %%
RNG_SEED = 42
torch.manual_seed(RNG_SEED)
np.random.seed(RNG_SEED)
data_type_torch = torch.float32


def groupby_formula(df, how="max"):
    """Group identical compositions together and preserve original indices.

    See https://stackoverflow.com/a/49216427/13697228

    Parameters
    ----------
    df : DataFrame
        At minimum should contain "formula" and "target" columns.
    how : str, optional
        How to perform the "groupby", either "mean" or "max", by default "max"

    Returns
    -------
    DataFrame
        The grouped DataFrame such that the original indices are preserved.
    """
    grp_df = (
        df.reset_index()
        .groupby(by="formula")
        .agg({"index": lambda x: tuple(x), "target": "max"})
        .reset_index()
    )
    return grp_df


def data(
    module,
    fname="train.csv",
    groupby=True,
    dummy=False,
    split=True,
    val_size=0.2,
    test_size=0.0,
    random_state=42,
):
    """Grab data from within the subdirectories (modules) of mat_discover.

    Parameters
    ----------
    module : Module
        The module within mat_discover that contains e.g. "train.csv". For example,
        from mat_discover.CrabNet.data.materials_data import elasticity
    fname : str, optional
        Filename of text file to open.
    dummy : bool, optional
        Whether to pare down the data to a small test set, by default False
    groupby : bool, optional
        Whether to use groupby_formula to filter identical compositions
    split : bool, optional
        Whether to split the data into train, val, and (optionally) test sets, by default True
    val_size : float, optional
        Validation dataset fraction, by default 0.2
    test_size : float, optional
        Test dataset fraction, by default 0.0
    random_state : int, optional
        seed to use for the train/val/test split, by default 42

    Returns
    -------
    DataFrame
        If split==False, then the full DataFrame is returned directly

    DataFrame, DataFrame
        If test_size == 0 and split==True, then training and validation DataFrames are returned.

    DataFrame, DataFrame, DataFrame
        If test_size > 0 and split==True, then training, validation, and test DataFrames are returned.
    """

    train_csv = open_text(module, fname)
    df = pd.read_csv(train_csv)

    if groupby:
        df = groupby_formula(df, how="max")

    if dummy:
        ntot = min(100, len(df))
        df = df.head(ntot)

    if split:
        if test_size > 0:
            df, test_df = train_test_split(
                df, test_size=test_size, random_state=random_state
            )

        train_df, val_df = train_test_split(
            df, test_size=val_size / (1 - test_size), random_state=random_state
        )

        if test_size > 0:
            return train_df, val_df, test_df
        else:
            return train_df, val_df
    else:
        return df


# %%
class Model:
    def __init__(
        self,
        model=None,
        model_name="UnnamedModel",
        n_elements="infer",
        verbose=True,
        force_cpu=False,
        prefer_last=True,
    ):
        if model is None:
            compute_device = get_compute_device(
                force_cpu=force_cpu, prefer_last=prefer_last
            )
            model = CrabNet(compute_device=compute_device).to(compute_device)
        self.model = model
        self.model_name = model_name
        self.data_loader = None
        self.train_loader = None
        self.classification = False
        self.n_elements = n_elements
        self.compute_device = model.compute_device
        self.fudge = 0.02  #  expected fractional tolerance (std. dev) ~= 2%
        self.verbose = verbose
        if self.verbose:
            print("\nModel architecture: out_dims, d_model, N, heads")
            print(
                f"{self.model.out_dims}, {self.model.d_model}, "
                f"{self.model.N}, {self.model.heads}"
            )
            print(f"Running on compute device: {self.compute_device}")
            print(f"Model size: {count_parameters(self.model)} parameters\n")

    def load_data(self, data, batch_size=2 ** 9, train=False):
        self.batch_size = batch_size
        inference = not train
        data_loaders = EDM_CsvLoader(
            data=data,
            batch_size=batch_size,
            n_elements=self.n_elements,
            inference=inference,
            verbose=self.verbose,
        )
        print(
            f"loading data with up to {data_loaders.n_elements:0.0f} "
            f"elements in the formula"
        )

        # update n_elements after loading dataset
        self.n_elements = data_loaders.n_elements

        data_loader = data_loaders.get_data_loaders(inference=inference)
        y = data_loader.dataset.data[1]
        if train:
            self.train_len = len(y)
            if self.classification:
                self.scaler = DummyScaler(y)
            else:
                self.scaler = Scaler(y)
            self.train_loader = data_loader
        self.data_loader = data_loader

    def train(self):
        self.model.train()
        ti = time()
        minima = []
        for i, data in enumerate(self.train_loader):
            X, y, formula = data
            y = self.scaler.scale(y)
            src, frac = X.squeeze(-1).chunk(2, dim=1)
            # add a small jitter to the input fractions to improve model
            # robustness and to increase stability
            # frac = frac * (1 + (torch.rand_like(frac)-0.5)*self.fudge)  # uniform
            frac = frac * (1 + (torch.randn_like(frac)) * self.fudge)  # normal
            frac = torch.clamp(frac, 0, 1)
            frac[src == 0] = 0
            frac = frac / frac.sum(dim=1).unsqueeze(1).repeat(1, frac.shape[-1])

            src = src.to(self.compute_device, dtype=torch.long, non_blocking=True)
            frac = frac.to(
                self.compute_device, dtype=data_type_torch, non_blocking=True
            )
            y = y.to(self.compute_device, dtype=data_type_torch, non_blocking=True)

            output = self.model.forward(src, frac)
            prediction, uncertainty = output.chunk(2, dim=-1)
            loss = self.criterion(prediction.view(-1), uncertainty.view(-1), y.view(-1))

            loss.backward()
            self.optimizer.step()
            self.optimizer.zero_grad()
            if self.stepping:
                self.lr_scheduler.step()

            swa_check = self.epochs_step * self.swa_start - 1
            epoch_check = (self.epoch + 1) % (2 * self.epochs_step) == 0
            learning_time = epoch_check and self.epoch >= swa_check
            if learning_time:
                act_v, pred_v, _, _ = self.predict(loader=self.data_loader)
                mae_v = mean_absolute_error(act_v, pred_v)
                self.optimizer.update_swa(mae_v)
                minima.append(self.optimizer.minimum_found)

        if learning_time and not any(minima):
            self.optimizer.discard_count += 1
            print(f"Epoch {self.epoch} failed to improve.")
            print(
                f"Discarded: {self.optimizer.discard_count}/"
                f"{self.discard_n} weight updates"
            )

        dt = time() - ti
        datalen = len(self.train_loader.dataset)
        # print(f'training speed: {datalen/dt:0.3f}')

    def fit(self, epochs=None, checkin=None, losscurve=False, learningcurve=True):
        assert_train_str = "Please Load Training Data (self.train_loader)"
        assert_val_str = "Please Load Validation Data (self.data_loader)"
        assert self.train_loader is not None, assert_train_str
        assert self.data_loader is not None, assert_val_str
        self.loss_curve = {}
        self.loss_curve["train"] = []
        self.loss_curve["val"] = []

        self.epochs_step = 10
        self.step_size = self.epochs_step * len(self.train_loader)
        print(
            f"stepping every {self.step_size} training passes,",
            f"cycling lr every {self.epochs_step} epochs",
        )
        if epochs is None:
            n_iterations = 1e4
            epochs = int(n_iterations / len(self.data_loader))
            print(f"running for {epochs} epochs")
        if checkin is None:
            checkin = self.epochs_step * 2
            print(f"checkin at {self.epochs_step*2} " f"epochs to match lr scheduler")
        if epochs % (self.epochs_step * 2) != 0:
            updated_epochs = epochs - epochs % (self.epochs_step * 2)
            print(
                f"epochs not divisible by {self.epochs_step * 2}, "
                f"updating epochs to {updated_epochs} for learning"
            )
            epochs = updated_epochs

        self.step_count = 0
        self.criterion = RobustL1
        if self.classification:
            print("Using BCE loss for classification task")
            self.criterion = BCEWithLogitsLoss
        base_optim = Lamb(params=self.model.parameters())
        optimizer = Lookahead(base_optimizer=base_optim)
        self.optimizer = SWA(optimizer)

        lr_scheduler = CyclicLR(
            self.optimizer,
            base_lr=1e-4,
            max_lr=6e-3,
            cycle_momentum=False,
            step_size_up=self.step_size,
        )

        self.swa_start = 2  # start at (n/2) cycle (lr minimum)
        self.lr_scheduler = lr_scheduler
        self.stepping = True
        self.lr_list = []
        self.xswa = []
        self.yswa = []
        self.discard_n = 3

        for epoch in range(epochs):
            self.epoch = epoch
            self.epochs = epochs
            ti = time()
            self.train()
            # print(f'epoch time: {(time() - ti):0.3f}')
            self.lr_list.append(self.optimizer.param_groups[0]["lr"])

            if (epoch + 1) % checkin == 0 or epoch == epochs - 1 or epoch == 0:
                ti = time()
                act_t, pred_t, _, _ = self.predict(loader=self.train_loader)
                dt = time() - ti
                datasize = len(act_t)
                # print(f'inference speed: {datasize/dt:0.3f}')
                mae_t = mean_absolute_error(act_t, pred_t)
                self.loss_curve["train"].append(mae_t)
                act_v, pred_v, _, _ = self.predict(loader=self.data_loader)
                mae_v = mean_absolute_error(act_v, pred_v)
                self.loss_curve["val"].append(mae_v)
                epoch_str = f"Epoch: {epoch}/{epochs} ---"
                train_str = f'train mae: {self.loss_curve["train"][-1]:0.3g}'
                val_str = f'val mae: {self.loss_curve["val"][-1]:0.3g}'
                if self.classification:
                    train_auc = roc_auc_score(act_t, pred_t)
                    val_auc = roc_auc_score(act_v, pred_v)
                    train_str = f"train auc: {train_auc:0.3f}"
                    val_str = f"val auc: {val_auc:0.3f}"
                print(epoch_str, train_str, val_str)

                if self.epoch >= (self.epochs_step * self.swa_start - 1):
                    if (self.epoch + 1) % (self.epochs_step * 2) == 0:
                        self.xswa.append(self.epoch)
                        self.yswa.append(mae_v)

                if losscurve:
                    plt.figure(figsize=(8, 5))
                    xval = np.arange(len(self.loss_curve["val"])) * checkin - 1
                    xval[0] = 0
                    plt.plot(xval, self.loss_curve["train"], "o-", label="train_mae")
                    plt.plot(xval, self.loss_curve["val"], "s--", label="val_mae")
                    plt.plot(
                        self.xswa, self.yswa, "o", ms=12, mfc="none", label="SWA point"
                    )
                    plt.ylim(0, 2 * np.mean(self.loss_curve["val"]))
                    plt.title(f"{self.model_name}")
                    plt.xlabel("epochs")
                    plt.ylabel("MAE")
                    plt.legend()
                    plt.show()

            if epoch == epochs - 1 or self.optimizer.discard_count >= self.discard_n:
                # save output df for stats tracking
                xval = np.arange(len(self.loss_curve["val"])) * checkin - 1
                xval[0] = 0
                tval = self.loss_curve["train"]
                vval = self.loss_curve["val"]
                os.makedirs("figures/lc_data", exist_ok=True)
                df_loss = pd.DataFrame([xval, tval, vval]).T
                df_loss.columns = ["epoch", "train loss", "val loss"]
                df_loss["swa"] = ["n"] * len(xval)
                df_loss.loc[df_loss["epoch"].isin(self.xswa), "swa"] = "y"
                df_loss.to_csv(f"figures/lc_data/{self.model_name}_lc.csv", index=False)

                # save output learning curve plot
                if learningcurve:
                    plt.figure(figsize=(8, 5))
                    xval = np.arange(len(self.loss_curve["val"])) * checkin - 1
                    xval[0] = 0
                    plt.plot(
                        xval,
                        self.loss_curve["train"],
                        "o-",
                        label="$\mathrm{MAE}_\mathrm{train}$",
                    )
                    plt.plot(
                        xval,
                        self.loss_curve["val"],
                        "s--",
                        label="$\mathrm{MAE}_\mathrm{val}$",
                    )
                    if self.epoch >= (self.epochs_step * self.swa_start - 1):
                        plt.plot(
                            self.xswa,
                            self.yswa,
                            "o",
                            ms=12,
                            mfc="none",
                            label="SWA point",
                        )
                    plt.ylim(0, 2 * np.mean(self.loss_curve["val"]))
                    plt.title(f"{self.model_name}")
                    plt.xlabel("epochs")
                    plt.ylabel("MAE")
                    plt.legend()
                    plt.savefig(f"figures/lc_data/{self.model_name}_lc.png")
                    plt.show()

            if self.optimizer.discard_count >= self.discard_n:
                print(
                    f"Discarded: {self.optimizer.discard_count}/"
                    f"{self.discard_n} weight updates, "
                    f"early-stopping now"
                )
                self.optimizer.swap_swa_sgd()
                break

        if not (self.optimizer.discard_count >= self.discard_n):
            self.optimizer.swap_swa_sgd()

    def predict(self, data=None, loader=None):
        if data is None and loader is None:
            raise SyntaxError("Specify either data *or* loader, not neither.")
        elif data is not None and loader is None:
            self.load_data(data)
            loader = self.data_loader
        elif data is not None and loader is not None:
            raise SyntaxError("Specify either data *or* loader, not both.")
        len_dataset = len(loader.dataset)
        n_atoms = int(len(loader.dataset[0][0]) / 2)
        act = np.zeros(len_dataset)
        pred = np.zeros(len_dataset)
        uncert = np.zeros(len_dataset)
        formulae = np.empty(len_dataset, dtype=list)
        atoms = np.empty((len_dataset, n_atoms))
        fractions = np.empty((len_dataset, n_atoms))
        self.model.eval()
        with torch.no_grad():
            for i, data in enumerate(loader):
                X, y, formula = data
                src, frac = X.squeeze(-1).chunk(2, dim=1)
                src = src.to(self.compute_device, dtype=torch.long, non_blocking=True)
                frac = frac.to(
                    self.compute_device, dtype=data_type_torch, non_blocking=True
                )
                y = y.to(self.compute_device, dtype=data_type_torch, non_blocking=True)
                output = self.model.forward(src, frac)
                prediction, uncertainty = output.chunk(2, dim=-1)
                uncertainty = torch.exp(uncertainty) * self.scaler.std
                prediction = self.scaler.unscale(prediction)
                if self.classification:
                    prediction = torch.sigmoid(prediction)

                data_loc = slice(i * self.batch_size, i * self.batch_size + len(y), 1)

                atoms[data_loc, :] = src.cpu().numpy()
                fractions[data_loc, :] = frac.cpu().numpy()
                act[data_loc] = y.view(-1).cpu().numpy()
                pred[data_loc] = prediction.view(-1).cpu().detach().numpy()
                uncert[data_loc] = uncertainty.view(-1).cpu().detach().numpy()
                formulae[data_loc] = formula

        return (act, pred, formulae, uncert)

    def save_network(self, model_name=None):
        if model_name is None:
            model_name = self.model_name
            os.makedirs("models/trained_models", exist_ok=True)
            path = f"models/trained_models/{model_name}.pth"
            print(f"Saving network ({model_name}) to {path}")
        else:
            path = f"models/trained_models/{model_name}.pth"
            print(f"Saving checkpoint ({model_name}) to {path}")

        self.network = {
            "weights": self.model.state_dict(),
            "scaler_state": self.scaler.state_dict(),
            "model_name": model_name,
        }
        torch.save(self.network, path)

    def load_network(self, model_data):
        if type(model_data) is str:
            path = f"models/trained_models/{model_data}"
            network = torch.load(path, map_location=self.compute_device)
        else:
            network = model_data
        base_optim = Lamb(params=self.model.parameters())
        optimizer = Lookahead(base_optimizer=base_optim)
        self.optimizer = SWA(optimizer)
        self.scaler = Scaler(torch.zeros(3))
        self.model.load_state_dict(network["weights"])
        self.scaler.load_state_dict(network["scaler_state"])
        self.model_name = network["model_name"]


# %%
if __name__ == "__main__":
    pass



import collections
import re
import pandas as pd
import numpy as np
import tqdm
import os

dirpath = os.getcwd()


class CompositionError(Exception):
    """Exception class for composition errors"""

    pass


def get_sym_dict(f, factor):
    sym_dict = collections.defaultdict(float)
    for m in re.finditer(r"([A-Z][a-z]*)\s*([-*\.\d]*)", f):
        el = m.group(1)
        amt = 1
        if m.group(2).strip() != "":
            amt = float(m.group(2))
        sym_dict[el] += amt * factor
        f = f.replace(m.group(), "", 1)
    if f.strip():
        raise CompositionError(f"{f} is an invalid formula!")
    return sym_dict


def parse_formula(formula):
    """
    Parameters
    ----------
        formula: str
            A string formula, e.g. Fe2O3, Li3Fe2(PO4)3.
    Return
    ----------
        sym_dict: dict
            A dictionary recording the composition of that formula.
    Notes
    ----------
        In the case of Metallofullerene formula (e.g. Y3N@C80),
        the @ mark will be dropped and passed to parser.
    """
    # for Metallofullerene like "Y3N@C80"
    formula = formula.replace("@", "")
    formula = formula.replace("[", "(")
    formula = formula.replace("]", ")")
    m = re.search(r"\(([^\(\)]+)\)\s*([\.\d]*)", formula)
    if m:
        factor = 1
        if m.group(2) != "":
            factor = float(m.group(2))
        unit_sym_dict = get_sym_dict(m.group(1), factor)
        expanded_sym = "".join(
            ["{}{}".format(el, amt) for el, amt in unit_sym_dict.items()]
        )
        expanded_formula = formula.replace(m.group(), expanded_sym)
        return parse_formula(expanded_formula)
    sym_dict = get_sym_dict(formula, 1)
    return sym_dict


def _fractional_composition(formula):
    elmap = parse_formula(formula)
    elamt = {}
    natoms = 0
    for k, v in elmap.items():
        if abs(v) >= 1e-6:
            elamt[k] = v
            natoms += abs(v)
    comp_frac = {key: elamt[key] / natoms for key in elamt}
    return comp_frac


def _fractional_composition_L(formula):
    comp_frac = _fractional_composition(formula)
    atoms = list(comp_frac.keys())
    counts = list(comp_frac.values())
    return atoms, counts


def _element_composition(formula):
    elmap = parse_formula(formula)
    elamt = {}
    natoms = 0
    for k, v in elmap.items():
        if abs(v) >= 1e-6:
            elamt[k] = v
            natoms += abs(v)
    return elamt


def _element_composition_L(formula):
    comp_frac = _element_composition(formula)
    atoms = list(comp_frac.keys())
    counts = list(comp_frac.values())
    return atoms, counts


def _assign_features(matrices, elem_info, formulae, sum_feat=False):
    formula_mat, count_mat, frac_mat, elem_mat, target_mat = matrices
    elem_symbols, elem_index, elem_missing = elem_info

    if sum_feat:
        sum_feats = []
    avg_feats = []
    range_feats = []
    # var_feats = []
    dev_feats = []
    max_feats = []
    min_feats = []
    mode_feats = []
    targets = []
    formulas = []
    skipped_formula = []

    for h in tqdm.tqdm(range(len(formulae)), desc="Assigning Features..."):
        elem_list = formula_mat[h]
        target = target_mat[h]
        formula = formulae[h]
        comp_mat = np.zeros(shape=(len(elem_list), elem_mat.shape[-1]))

        for i, elem in enumerate(elem_list):
            if elem in elem_missing:
                skipped_formula.append(formula)
            else:
                row = elem_index[elem_symbols.index(elem)]
                comp_mat[i, :] = elem_mat[row]

        range_feats.append(np.ptp(comp_mat, axis=0))
        # var_feats.append(comp_mat.var(axis=0))
        max_feats.append(comp_mat.max(axis=0))
        min_feats.append(comp_mat.min(axis=0))

        comp_frac_mat = comp_mat.T * frac_mat[h]
        comp_frac_mat = comp_frac_mat.T
        avg_feats.append(comp_frac_mat.sum(axis=0))

        dev = np.abs(comp_mat - comp_frac_mat.sum(axis=0))
        dev = dev.T * frac_mat[h]
        dev = dev.T.sum(axis=0)
        dev_feats.append(dev)

        prominant = np.isclose(frac_mat[h], max(frac_mat[h]))
        mode = comp_mat[prominant].min(axis=0)
        mode_feats.append(mode)

        comp_sum_mat = comp_mat.T * count_mat[h]
        comp_sum_mat = comp_sum_mat.T
        if sum_feat:
            sum_feats.append(comp_sum_mat.sum(axis=0))

        targets.append(target)
        formulas.append(formula)

    if len(skipped_formula) > 0:
        print(
            "\nNOTE: Your data contains formula with exotic elements.",
            "These were skipped.",
        )
    if sum_feat:
        conc_list = [
            sum_feats,
            avg_feats,
            dev_feats,
            range_feats,
            max_feats,
            min_feats,
            mode_feats,
        ]
        feats = np.concatenate(conc_list, axis=1)
    else:
        conc_list = [
            avg_feats,
            dev_feats,
            range_feats,
            max_feats,
            min_feats,
            mode_feats,
        ]
        feats = np.concatenate(conc_list, axis=1)

    return feats, targets, formulas, skipped_formula


def generate_features(
    df,
    elem_prop="oliynyk",
    drop_duplicates=False,
    extend_features=False,
    sum_feat=False,
    mini=False,
):
    """
    Parameters
    ----------
    df: Pandas.DataFrame()
        X column dataframe of form:
            df.columns.values = array(['formula', 'target',
                                       'extended1', 'extended2', ...],
                                      dtype=object)
    elem_prop: str
        valid element properties:
            'oliynyk',
            'jarvis',
            'atom2vec',
            'magpie',
            'mat2vec',
            'onehot'
    drop_duplicates: boolean
        Decide to keep or drop duplicate compositions
    extend_features: boolean
        Decide whether to use non ["formula", "target"] columns as additional
        features.
    Return
    ----------
    X: pd.DataFrame()
        Feature Matrix with NaN values filled using the median feature value
        for dataset
    y: pd.Series()
        Target values
    formulae: pd.Series()
        Formula associated with X and y
    """
    if drop_duplicates:
        if df["formula"].value_counts()[0] > 1:
            df.drop_duplicates("formula", inplace=True)
            print("Duplicate formula(e) removed using default pandas function")

    # fmt: off
    all_symbols = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na',
                   'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc',
                   'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga',
                   'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb',
                   'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb',
                   'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm',
                   'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu',
                   'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', 'Tl',
                   'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa',
                   'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md',
                   'No', 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg',
                   'Cn', 'Nh', 'Fl', 'Mc', 'Lv', 'Ts', 'Og']
    # fmt: on

    elem_props = pd.read_csv(dirpath + "/data/element_properties/" + elem_prop + ".csv")
    elem_props.index = elem_props["element"].values
    elem_props.drop(["element"], inplace=True, axis=1)

    elem_symbols = elem_props.index.tolist()
    elem_index = np.arange(0, elem_props.shape[0], 1)
    elem_missing = list(set(all_symbols) - set(elem_symbols))

    elem_props_columns = elem_props.columns.values

    column_names = np.concatenate(
        [
            "avg_" + elem_props_columns,
            "dev_" + elem_props_columns,
            "range_" + elem_props_columns,
            "max_" + elem_props_columns,
            "min_" + elem_props_columns,
            "mode_" + elem_props_columns,
        ]
    )
    if sum_feat:
        column_names = np.concatenate(["sum_" + elem_props_columns, column_names])

    # make empty list where we will store the property value
    targets = []
    # store formula
    formulae = []
    # add the values to the list using a for loop

    elem_mat = elem_props.values

    formula_mat = []
    count_mat = []
    frac_mat = []
    target_mat = []

    if extend_features:
        features = df.columns.values.tolist()
        features.remove("target")
        extra_features = df[features]

    for index in tqdm.tqdm(df.index.values, desc="Processing Input Data"):
        formula, target = df.loc[index, "formula"], df.loc[index, "target"]
        if "x" in formula:
            continue
        l1, l2 = _element_composition_L(formula)
        formula_mat.append(l1)
        count_mat.append(l2)
        _, l3 = _fractional_composition_L(formula)
        frac_mat.append(l3)
        target_mat.append(target)
        formulae.append(formula)

    print("\tfeaturizing compositions...".title())

    matrices = [formula_mat, count_mat, frac_mat, elem_mat, target_mat]
    elem_info = [elem_symbols, elem_index, elem_missing]
    feats, targets, formulae, skipped = _assign_features(
        matrices, elem_info, formulae, sum_feat=sum_feat
    )

    print("\tcreating pandas objects...".title())

    # split feature vectors and targets as X and y
    X = pd.DataFrame(feats, columns=column_names, index=formulae)
    y = pd.Series(targets, index=formulae, name="target")
    formulae = pd.Series(formulae, index=formulae, name="formula")
    if extend_features:
        extended = pd.DataFrame(extra_features, columns=features)
        extended = extended.set_index("formula", drop=True)
        X = pd.concat([X, extended], axis=1)

    # reset dataframe indices
    X.reset_index(drop=True, inplace=True)
    y.reset_index(drop=True, inplace=True)
    formulae.reset_index(drop=True, inplace=True)

    # drop elements that aren't included in the elmenetal properties list.
    # These will be returned as feature rows completely full of NaN values.
    X.dropna(inplace=True, how="all")
    y = y.iloc[X.index]
    formulae = formulae.iloc[X.index]

    # get the column names
    cols = X.columns.values
    # find the median value of each column
    median_values = X[cols].median()
    # fill the missing values in each column with the column's median value
    X[cols] = X[cols].fillna(median_values)

    # Only return the avg/sum of element properties.

    if mini:
        np.random.seed(42)
        booleans = np.random.rand(X.shape[-1]) <= 64 / X.shape[-1]
        X = X.iloc[:, booleans]

    return X, y, formulae, skipped


# if __name__ == '__main__':
#     pass



"""Train CrabNet using data for a given material property.

Use the crabnet environment.
"""
import os
from os.path import exists, join, dirname

# from warnings import warn
import numpy as np
import pandas as pd
import torch

from sklearn.metrics import roc_auc_score

from crabnet.kingcrab import CrabNet  # type: ignore
from crabnet.model import Model  # type: ignore

from crabnet.utils.get_compute_device import get_compute_device

RNG_SEED = 42
torch.manual_seed(RNG_SEED)
np.random.seed(RNG_SEED)


# %%
def get_model(
    data_dir=join(dirname(__file__), "data", "materials_data"),
    mat_prop=None,
    train_df=None,
    val_df=None,
    test_df=None,
    classification=False,
    batch_size=None,
    transfer=None,
    verbose=True,
    losscurve=False,
    learningcurve=True,
    force_cpu=False,
    prefer_last=True,
):
    compute_device = get_compute_device(force_cpu=force_cpu, prefer_last=prefer_last)
    if train_df is None and val_df is None and test_df is None:
        if mat_prop is None:
            mat_prop = "example_materials_property"
        use_path = True
    else:
        if mat_prop is None:
            mat_prop = "DataFrame_property"
        use_path = False

    # Get the TorchedCrabNet architecture loaded
    model = Model(
        CrabNet(compute_device=compute_device).to(compute_device),
        model_name=f"{mat_prop}",
        verbose=verbose,
    )

    # Train network starting at pretrained weights
    if transfer is not None:
        model.load_network(f"{transfer}.pth")
        model.model_name = f"{mat_prop}"

    # Apply BCEWithLogitsLoss to model output if binary classification is True
    if classification:
        model.classification = True

    if use_path:
        # Get the datafiles you will learn from
        train_data = f"{data_dir}/{mat_prop}/train.csv"
        try:
            val_data = f"{data_dir}/{mat_prop}/val.csv"
        except IOError:
            print(
                "Please ensure you have train (train.csv) and validation data",
                f'(val.csv) in folder "data/materials_data/{mat_prop}"',
            )
        data_size = pd.read_csv(train_data).shape[0]
    else:
        train_data = train_df
        val_data = val_df
        data_size = train_data.shape[0]

    # Load the train and validation data before fitting the network
    batch_size = 2 ** round(np.log2(data_size) - 4)
    if batch_size < 2 ** 7:
        batch_size = 2 ** 7
    if batch_size > 2 ** 12:
        batch_size = 2 ** 12
    model.load_data(train_data, batch_size=batch_size, train=True)
    print(
        f"training with batchsize {model.batch_size} "
        f"(2**{np.log2(model.batch_size):0.3f})"
    )
    if val_data is not None:
        model.load_data(val_data, batch_size=batch_size)

    # Set the number of epochs, decide if you want a loss curve to be plotted
    model.fit(epochs=40, losscurve=losscurve, learningcurve=learningcurve)

    # Save the network (saved as f"{model_name}.pth")
    model.save_network()
    return model


def to_csv(output, save_name):
    # parse output and save to csv
    act, pred, formulae, uncertainty = output
    df = pd.DataFrame([formulae, act, pred, uncertainty]).T
    df.columns = ["composition", "target", "pred-0", "uncertainty"]
    if save_name is not None:
        save_path = "model_predictions"
        os.makedirs(save_path, exist_ok=True)
        df.to_csv(f"{save_path}/{save_name}", index_label="Index")
    return df


def load_model(model, mat_prop, classification, data, verbose=True):
    # Load up a saved network.
    if type(model) is str:
        usepath = True
        model = Model(
            CrabNet(compute_device=compute_device).to(compute_device),
            model_name=f"{mat_prop}",
            verbose=verbose,
        )
        model_path = f"{mat_prop}.pth"
        model.load_network(model_path)
    else:
        usepath = False

    # Check if classifcation task
    if classification:
        model.classification = True

    # Load the data you want to predict with

    if usepath and type(data) is str:
        data = f"{mat_prop}/{mat_prop}/{data}"
    # data is reloaded to model.data_loader
    model.load_data(data, batch_size=2 ** 9, train=False)
    return model


def get_results(model):
    output = model.predict(loader=model.data_loader)  # predict the data saved here
    return model, output


def save_results(model, mat_prop, classification, data, verbose=True):
    if type(model) is str:
        usepath = True
        model = load_model(model, mat_prop, classification, data, verbose=verbose)
    else:
        usepath = False
        model.load_data(data, batch_size=2 ** 9, train=False)
    model, output = get_results(model)

    # Get appropriate metrics for saving to csv
    if model.classification:
        auc = roc_auc_score(output[0], output[1])
        print(f"{mat_prop} ROC AUC: {auc:0.3f}")
    else:
        mae = np.abs(output[0] - output[1]).mean()
        print(f"{mat_prop} mae: {mae:0.3g}")

    # save predictions to a csv
    if usepath:
        fname = f'{mat_prop}_{data.replace(".csv", "")}_output.csv'
    else:
        fname = None
    df = to_csv(output, fname)
    return model, mae, df


def main(
    train_df=None,
    val_df=None,
    test_df=None,
    data_dir=join(dirname(__file__), "data", "materials_data"),
    mat_prop=None,
    classification=False,
    train=True,
    losscurve=False,
    learningcurve=True,
    verbose=True,
):
    """
    Train CrabNet model, save predictions, and output mean absolute error.

    Parameters
    ----------
    data_dir : TYPE, optional
        Directory where your data is stored.
        The default is 'data/materials_data'.
    mat_prop : TYPE, optional
        Folder with your materials properties.
        The default is 'example_materials_property'.
    classification : bool, optional
        Whether to perform classification. Performs regression if False.
        The default is False.
    train : bool, optional
        Whether to perform training. The default is True.

    Returns
    -------
    None.

    """
    if train_df is None and val_df is None and test_df is None:
        if mat_prop is None:
            mat_prop = "example_materials_property"
        train_data = join(data_dir, mat_prop, "train.csv")
        val_data = join(data_dir, mat_prop, "val.csv")
        test_data = join(data_dir, mat_prop, "test.csv")
        use_test = exists(test_data)
    else:
        mat_prop = "DataFrame_property"
        train_data = train_df
        val_data = val_df
        test_data = test_df
        use_test = test_df is not None

    # Train your model using the "get_model" function
    if train:
        if verbose:
            print(f'Property "{mat_prop}" selected for training')
        model = get_model(
            data_dir,
            mat_prop=mat_prop,
            classification=classification,
            verbose=verbose,
            train_df=train_df,
            val_df=val_df,
            test_df=test_df,
            losscurve=losscurve,
            learningcurve=learningcurve,
        )

    if verbose:
        cutter = "====================================================="
        first = " " * ((len(cutter) - len(mat_prop)) // 2) + " " * int(
            (len(mat_prop) + 1) % 2
        )
        last = " " * ((len(cutter) - len(mat_prop)) // 2)
        print("=====================================================")
        print(f"{first}{mat_prop}{last}")
        print("=====================================================")
        print("calculating train mae")
    model_train, mae_train, train_pred_df = save_results(
        model, mat_prop, classification, train_data, verbose=False
    )
    if val_data is not None:
        if verbose:
            print("-----------------------------------------------------")
            print("calculating val mae")
        model_val, mae_val, val_pred_df = save_results(
            model, mat_prop, classification, val_data, verbose=False
        )
    if use_test:
        if verbose:
            print("-----------------------------------------------------")
            print("calculating test mae")
        model_test, mae_test, test_pred_df = save_results(
            model, mat_prop, classification, test_data, verbose=False
        )
    else:
        test_pred_df = None
    if verbose:
        print("=====================================================")
    return train_pred_df, val_pred_df, test_pred_df


# %%
if __name__ == "__main__":
    main()

# %% Code Graveyard
# HACK:
# try:
#     from .crabnet.kingcrab import CrabNet
#     from .crabnet.model import Model
# except ImportWarning:
#     warn(
#         "relative import didn't work, probably because code is being executed as script instead of package",
#         ImportWarning,
#     )



import torch


# %%
def get_compute_device(force_cpu=False, prefer_last=True):
    CUDA_available = torch.cuda.is_available()
    if CUDA_available:
        CUDA_count = torch.cuda.device_count()

    compute_device = torch.device("cpu")

    if force_cpu or not CUDA_available:
        return compute_device

    elif prefer_last and CUDA_count > 1:
        compute_device = torch.device(f"cuda:{CUDA_count - 1}")
        return compute_device

    else:
        compute_device = torch.device("cuda")
        return torch.device("cuda:0")



import os
import numpy as np
import pandas as pd

from time import time
import matplotlib.pyplot as plt

from ast import literal_eval

from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import r2_score, mean_absolute_error, mean_squared_error

from joblib import dump, load

from crabnet.utils.utils import get_cbfv
from crabnet.utils.utils import CONSTANTS

plt.rcParams.update({"font.size": 16})


# %%
cons = CONSTANTS()

bm_names = cons.benchmark_props
bm_pretty = cons.benchmark_names_dict
mb_names = cons.matbench_props
mb_pretty = cons.matbench_names_dict
elem_props = cons.eps
elem_props = ["magpie"]

mat_props = mb_names + bm_names

model_names = {
    "RandomForestRegressor": RandomForestRegressor,
}


# %%
def concatenate_best_results(directory):
    files = os.listdir(directory)

    csv_list = [file for file in files if ".csv" in file if "best_" in file]
    csv_paths = [os.path.join(directory, file) for file in csv_list]

    df = pd.DataFrame()

    for file in csv_paths:
        df_file = pd.read_csv(file)
        df = pd.concat([df, df_file], axis=0, ignore_index=True)

    df = df[
        [
            "estimator",
            "elem_prop",
            "mat_prop",
            "params",
            "mean_test_neg_MAE",
            "mean_test_r2",
        ]
    ]

    return df


def get_best_model(df, mat_prop, sortby="mean_test_neg_MAE"):
    df_mp = df[df["mat_prop"] == mat_prop]
    df_best_row = df_mp.sort_values(sortby, axis=0, ascending=False).iloc[0]

    params = df_best_row["params"]
    params = literal_eval(params)

    est_name = df_best_row["estimator"]
    est_class = model_names[est_name]
    estimator = est_class(**params)

    elem_prop = df_best_row["elem_prop"]

    return est_name, estimator, params, elem_prop


# %%
if __name__ == "__main__":
    metrics_path = r"metrics/rf_gridsearch/"

    bm_predictions_path = r"metrics/rf_benchmark__predictions/"
    mb_predictions_path = r"metrics/rf_matbench__predictions/"
    clf_save_path = f"models/trained_rf/"
    os.makedirs(clf_save_path, exist_ok=True)
    os.makedirs(metrics_path, exist_ok=True)
    os.makedirs(bm_predictions_path, exist_ok=True)
    os.makedirs(mb_predictions_path, exist_ok=True)

    bm_props_dir = r"data/benchmark_data/"
    mb_props_dir = r"data/matbench_cv/"

    df_best_classics = concatenate_best_results(metrics_path)

    retrain_columns = ["estimator_name", "estimator", "params", "mat_prop", "elem_prop"]
    df_models_to_retrain = pd.DataFrame(columns=retrain_columns)

    for mp in mat_props:
        output = get_best_model(df_best_classics, mp)
        est_name, estimator, params, elem_prop = output

        row_dict = {
            "estimator_name": est_name,
            "estimator": estimator,
            "params": params,
            "mat_prop": mp,
            "elem_prop": elem_prop,
        }
        df_models_to_retrain = df_models_to_retrain.append(row_dict, ignore_index=True)

    df_retrain_columns = [
        "estimator_name",
        "fit_time",
        "mae_train",
        "mse_train",
        "r2_train",
        "mae_test",
        "mse_test",
        "r2_test",
        "params",
        "mat_prop",
        "elem_prop",
    ]
    df_retrain_results = pd.DataFrame(columns=df_retrain_columns)
    ti_retrain_classics = time()

    mb_cvs = 5
    for i in range(len(df_models_to_retrain)):
        output = df_models_to_retrain.iloc[i].values
        est_name, estimator, param, mat_prop, elem_prop = output

        if mat_prop in bm_names:
            df_predictions = pd.DataFrame()
            trainpath = os.path.join(bm_props_dir, mat_prop, "train.csv")
            valpath = os.path.join(bm_props_dir, mat_prop, "val.csv")
            testpath = os.path.join(bm_props_dir, mat_prop, "test.csv")

            X, y, form, skipped = get_cbfv(trainpath, elem_prop=elem_prop)
            X_val, y_val, form_val, skipped_val = get_cbfv(valpath, elem_prop=elem_prop)
            X_test, y_test, form_test, skipped_test = get_cbfv(
                testpath, elem_prop=elem_prop
            )

            X = pd.concat([X, X_val], axis=0, ignore_index=True)
            y = pd.concat([y, y_val], axis=0, ignore_index=True)
            form = pd.concat([form, form_val], axis=0, ignore_index=True)

            print(f"fitting {mat_prop} with {elem_prop} using {est_name}")
            if X.shape[0] > 2000:
                print(
                    f"there are {X.shape[0]} data samples, " f"this may take a while..."
                )
            ti_model = time()
            estimator.fit(X, y)
            dt_model = time() - ti_model
            print(f"finished fitting {mat_prop} with {elem_prop} using {est_name}")
            print(f"fitting time: {dt_model:0.4f} s")

            savepath = f"models/trained_rf/{mat_prop}.joblib"
            print(f"Saving rf model {savepath}")
            dump(estimator, savepath, compress=3, protocol=5)

            print(f"evaluating fitted model on whole trainset and testset")
            ti_eval = time()
            target_test = y_test
            pred_test = estimator.predict(X_test)

            target_train = y
            pred_train = estimator.predict(X)

            r2_test = r2_score(target_test, pred_test)
            mae_test = mean_absolute_error(target_test, pred_test)
            mse_test = mean_squared_error(target_test, pred_test)

            r2_train = r2_score(target_train, pred_train)
            mae_train = mean_absolute_error(target_train, pred_train)
            mse_train = mean_squared_error(target_train, pred_train)

            dt_eval = time() - ti_eval
            print(f"evaluation time: {dt_eval:0.4f} s")

            df_retrain_row = {
                "estimator_name": est_name,
                "fit_time": dt_model,
                "mae_train": mae_train,
                "mse_train": mse_train,
                "r2_train": r2_train,
                "mae_test": mae_test,
                "mse_test": mse_test,
                "r2_test": r2_test,
                "params": param,
                "mat_prop": mat_prop,
                "cv": "NaN",
                "elem_prop": elem_prop,
            }

            df_retrain_results = df_retrain_results.append(
                df_retrain_row, ignore_index=True
            )

            df_predictions["composition"] = form_test
            df_predictions["target"] = target_test
            df_predictions["pred"] = pred_test
            outfile = os.path.join(bm_predictions_path, f"{mat_prop}_test_output.csv")
            df_predictions.to_csv(outfile)
            print(f"mae_test for {mat_prop}: {mae_test}")

        elif mat_prop in mb_names:
            trainpath_pre = os.path.join(mb_props_dir, mat_prop, "train")
            valpath_pre = os.path.join(mb_props_dir, mat_prop, "val")
            testpath_pre = os.path.join(mb_props_dir, mat_prop, "test")

            maes = []
            for cv in range(mb_cvs):
                df_predictions = pd.DataFrame()
                trainpath = f"{trainpath_pre}{cv}.csv"
                valpath = f"{valpath_pre}{cv}.csv"
                testpath = f"{testpath_pre}{cv}.csv"

                X, y, form, skipped = get_cbfv(trainpath, elem_prop=elem_prop)
                X_val, y_val, form_val, skipped_val = get_cbfv(
                    valpath, elem_prop=elem_prop
                )
                X_test, y_test, form_test, skipped_test = get_cbfv(
                    testpath, elem_prop=elem_prop
                )

                X = pd.concat([X, X_val], axis=0, ignore_index=True)
                y = pd.concat([y, y_val], axis=0, ignore_index=True)
                form = pd.concat([form, form_val], axis=0, ignore_index=True)

                print(f"fitting {mat_prop} with {elem_prop} using {est_name}")
                print(f"cv {cv}")
                if X.shape[0] > 2000:
                    print(
                        f"there are {X.shape[0]} data samples, "
                        f"this may take a while..."
                    )
                ti_model = time()
                estimator.fit(X, y)
                dt_model = time() - ti_model
                print(f"finished fitting {mat_prop} with {elem_prop} using {est_name}")
                print(f"fitting time: {dt_model:0.4f} s")

                savepath = f"models/trained_rf/{mat_prop}_cv{cv}.joblib"
                print(f"Saving rf model {savepath}")
                dump(estimator, savepath, compress=3, protocol=5)

                print(f"evaluating fitted model on whole trainset and testset")
                ti_eval = time()
                target_test = y_test
                pred_test = estimator.predict(X_test)

                target_train = y
                pred_train = estimator.predict(X)

                r2_test = r2_score(target_test, pred_test)
                mae_test = mean_absolute_error(target_test, pred_test)
                mse_test = mean_squared_error(target_test, pred_test)

                r2_train = r2_score(target_train, pred_train)
                mae_train = mean_absolute_error(target_train, pred_train)
                mse_train = mean_squared_error(target_train, pred_train)

                dt_eval = time() - ti_eval
                print(f"evaluation time: {dt_eval:0.4f} s")

                df_retrain_row = {
                    "estimator_name": est_name,
                    "fit_time": dt_model,
                    "mae_train": mae_train,
                    "mse_train": mse_train,
                    "r2_train": r2_train,
                    "mae_test": mae_test,
                    "mse_test": mse_test,
                    "r2_test": r2_test,
                    "params": param,
                    "mat_prop": mat_prop,
                    "cv": cv,
                    "elem_prop": elem_prop,
                }

                df_retrain_results = df_retrain_results.append(
                    df_retrain_row, ignore_index=True
                )

                df_predictions["composition"] = form_test
                df_predictions["target"] = target_test
                df_predictions["pred"] = pred_test
                outfile = os.path.join(
                    mb_predictions_path, f"{mat_prop}_test_output_cv{cv}.csv"
                )
                df_predictions.to_csv(outfile)
                print(f"mae_test: {mae_test}")
                maes.append(mae_test)
            print(f"average mae_test for {mat_prop}: {np.mean(maes)}")

    dt_retrain_classics = time() - ti_retrain_classics
    print("*********** retrain_classics finished ***********")
    print(f"retrain_classics finished, elapsed time: " f"{dt_retrain_classics:0.4g} s")
    print("*********** retrain_classics finished ***********")

    outfile = os.path.join(metrics_path, "retrained_test_scores.csv")
    df_retrain_results.to_csv(outfile, index=False)
